import numpy as np
import pytest

import nengo.dists as dists
import nengo.utils.numpy as npext
from nengo.exceptions import ValidationError


def test_pdf(rng):
    s = 0.25
    f = lambda x: (np.exp(-0.5 * (x + 0.5)**2 / s**2) +
                   np.exp(-0.5 * (x - 0.5)**2 / s**2))

    xref = np.linspace(-2, 2, 101)
    pref = f(xref)
    pref /= pref.sum()
    dist = dists.PDF(xref, pref)

    n = 100000
    samples = dist.sample(n, rng=rng)
    h, xedges = np.histogram(samples, bins=101)
    x = 0.5 * (xedges[:-1] + xedges[1:])
    dx = np.diff(xedges)
    y = h / float(h.sum()) / dx
    z = f(x)
    z = z / z.sum() / dx
    assert np.allclose(y, z, atol=0.05)


@pytest.mark.parametrize("low,high", [(-2, -1), (-1, 1), (1, 2), (1, -1)])
def test_uniform(low, high, rng):
    n = 100
    dist = dists.Uniform(low, high)
    samples = dist.sample(n, rng=rng)
    if low < high:
        assert np.all(samples >= low)
        assert np.all(samples < high)
    else:
        assert np.all(samples <= low)
        assert np.all(samples > high)
    hist, _ = np.histogram(samples, bins=5)
    assert np.allclose(hist - np.mean(hist), 0, atol=0.1 * n)


@pytest.mark.parametrize("mean,std", [(0, 1), (0, 0), (10, 2)])
def test_gaussian(mean, std, rng):
    n = 100
    if std <= 0:
        with pytest.raises(ValueError):
            dist = dists.Gaussian(mean, std)
    else:
        dist = dists.Gaussian(mean, std)
        samples = dist.sample(n, rng=rng)
        assert abs(np.mean(samples) - mean) < 2 * std / np.sqrt(n)
        assert abs(np.std(samples) - std) < 0.25  # using chi2 for n=100


@pytest.mark.parametrize("scale,shift,high", [
    (1., 0., np.inf), (10., 0., 1.), (0.1, 0.3, 1.0)])
def test_exponential(scale, shift, high, rng):
    n = 100
    dist = dists.Exponential(scale, shift=shift, high=high)
    samples = dist.sample(n, rng=rng)
    assert np.all(samples >= shift)
    assert np.all(samples <= high)
    # approximation of 95% confidence interval
    ci = scale * 1.96 / np.sqrt(n)
    if scale + ci < high:
        assert abs(np.mean(samples - shift) - scale) < ci


@pytest.mark.parametrize(
    "min_magnitude,d", [(0, 1), (0, 2), (0, 5), (0.6, 1), (0.3, 2), (0.4, 5)])
def test_hypersphere_volume(min_magnitude, d, rng):
    n = 150 * d
    dist = dists.UniformHypersphere(min_magnitude=min_magnitude)
    samples = dist.sample(n, d, rng=rng)
    assert samples.shape == (n, d)
    assert np.allclose(np.mean(samples, axis=0), 0, atol=0.1)
    assert np.all(npext.norm(samples, axis=1) >= min_magnitude)


@pytest.mark.parametrize("dimensions", [1, 2, 5])
def test_hypersphere_surface(dimensions, rng):
    n = 150 * dimensions
    dist = dists.UniformHypersphere(surface=True)
    samples = dist.sample(n, dimensions, rng=rng)
    assert samples.shape == (n, dimensions)
    assert np.allclose(npext.norm(samples, axis=1), 1)
    assert np.allclose(np.mean(samples, axis=0), 0, atol=0.25 / dimensions)


def test_hypersphere_dimension_fail(rng):
    with pytest.raises(ValueError):
        dists.UniformHypersphere(0).sample(1, 0)


def test_hypersphere_warns(rng):
    with pytest.warns(UserWarning):
        dists.UniformHypersphere(surface=True, min_magnitude=0.1)


@pytest.mark.parametrize("weights", [None, [5, 1, 2, 9], [3, 2, 1, 0]])
def test_choice(weights, rng):
    n = 1000
    choices = [[1, 1], [1, -1], [-1, 1], [-1, -1]]
    N = len(choices)

    dist = dists.Choice(choices, weights=weights)
    # If d is passed, it has to match
    with pytest.raises(ValueError):
        dist.sample(n, d=4, rng=rng)
    sample = dist.sample(n, rng=rng)
    tsample, tchoices = list(map(tuple, sample)), list(map(tuple, choices))

    # check that frequency of choices matches weights
    inds = [tchoices.index(s) for s in tsample]
    hist, bins = np.histogram(inds, bins=np.linspace(-0.5, N - 0.5, N + 1))
    p_empirical = hist / float(hist.sum())
    p = np.ones(N) / N if dist.p is None else dist.p
    sterr = 1. / np.sqrt(n)  # expected maximum standard error
    assert np.allclose(p, p_empirical, atol=2 * sterr)


@pytest.mark.parametrize("shape", [(12, 2), (7, 1), (7,), (1, 1)])
def test_samples(shape, rng):
    samples = rng.random_sample(size=shape)
    d = dists.Samples(samples)
    dims = None if len(shape) == 1 else shape[1]
    assert np.allclose(d.sample(shape[0], dims), samples)


@pytest.mark.parametrize("samples", [[1., 2., 3.], [[1, 2], [3, 4]]])
def test_samples_list(samples):
    d = dists.Samples(samples)
    shape = np.array(samples).shape
    dims = None if len(shape) == 1 else shape[1]
    assert np.allclose(d.sample(shape[0], dims), samples)


def test_samples_errors(rng):
    samples = rng.random_sample(size=(12, 2))
    with pytest.raises(ValidationError):
        dists.Samples(samples).sample(11, 2)
    with pytest.raises(ValidationError):
        dists.Samples(samples).sample(12, 1)
    with pytest.raises(ValidationError):
        dists.Samples(samples).sample(12)

    samples = rng.random_sample(size=12)
    with pytest.raises(ValidationError):
        dists.Samples(samples).sample(12, 2)


@pytest.mark.parametrize("n,m", [(99, 1), (50, 50)])
def test_sqrt_beta(n, m, rng):
    num_samples = 250
    num_bins = 5

    vectors = rng.randn(num_samples, n + m)
    vectors /= npext.norm(vectors, axis=1, keepdims=True)
    expectation, _ = np.histogram(
        npext.norm(vectors[:, :m], axis=1), bins=num_bins)

    dist = dists.SqrtBeta(n, m)
    samples = dist.sample(num_samples, 1, rng=rng)
    hist, _ = np.histogram(samples, bins=num_bins)

    assert np.all(np.abs(np.asfarray(hist - expectation) / num_samples) < 0.16)


@pytest.mark.parametrize("n,m", [(4, 1), (10, 5)])
def test_sqrt_beta_analytical(n, m, rng):
    """Tests pdf, cdf, and ppf of SqrtBeta distribution."""
    pytest.importorskip('scipy')  # beta and betainc

    dt = 0.001
    x = np.arange(dt, 1+dt, dt)

    dist = dists.SqrtBeta(n, m)

    pdf = dist.pdf(x)
    cdf = dist.cdf(x)
    ppf = dist.ppf(cdf)

    # The pdf should reflect the samples
    num_samples = 2500
    num_bins = 5

    samples = dist.sample(num_samples, rng=rng)
    act_hist, _ = np.histogram(samples, bins=num_bins)
    bin_points = np.linspace(0, 1, num_bins+1)
    bin_cdf = dist.cdf(bin_points)
    exp_freq = bin_cdf[1:] - bin_cdf[:-1]
    assert np.all(np.abs(np.asfarray(act_hist) / num_samples - exp_freq) < 0.1)

    # The cdf should be the accumulated pdf
    assert np.allclose(cdf, np.cumsum(pdf) * dt, atol=0.01)

    # The ppf should give back x
    assert np.allclose(x, ppf, atol=0.01)


@pytest.mark.parametrize("d", [2, 3, 10, 50])
def test_cosine_similarity(d, rng):
    """Tests CosineSimilarity sampling."""
    num_samples = 2500
    num_bins = 5

    # Check that it gives a single dimension from UniformHypersphere
    exp_dist = dists.UniformHypersphere(surface=True)
    act_dist = dists.CosineSimilarity(d)

    exp = exp_dist.sample(num_samples, d, rng=rng)[:, 0]
    act = act_dist.sample(num_samples, rng=rng)

    exp_hist, _ = np.histogram(exp, bins=num_bins)
    act_hist, _ = np.histogram(act, bins=num_bins)

    assert np.all(np.abs(np.asfarray(exp_hist - act_hist) / num_samples) < 0.1)


@pytest.mark.parametrize("d", [2, 3, 10])
def test_cosine_analytical(d):
    pytest.importorskip('scipy')  # beta, betainc, betaincinv

    dt = 0.0001
    x = np.arange(-1+dt, 1, dt)

    def p(x, d):
        # unnormalized CosineSimilarity distribution, derived by Eric H.
        return (1 - x*x)**((d - 3) / 2.0)

    dist = dists.CosineSimilarity(d)

    pdf_exp = dist.pdf(x)
    pdf_act = p(x, d)

    cdf_exp = dist.cdf(x)
    cdf_act = np.cumsum(pdf_act) / np.sum(pdf_act)

    # Check that we get the expected pdf after normalization
    assert np.allclose(
        pdf_exp / np.sum(pdf_exp), pdf_act / np.sum(pdf_act), atol=0.01)

    # Check that this accumulates to the expected cdf
    assert np.allclose(cdf_exp, cdf_act, atol=0.01)

    # Check that the inverse cdf gives back x
    assert np.allclose(dist.ppf(cdf_exp), x, atol=0.01)


def test_cosine_sample_shape(seed):
    """"Tests that CosineSimilarity sample has correct shape."""
    # sampling (n, d) should be the exact same as sampling (n*d,)
    n = 3
    d = 4
    dist = dists.CosineSimilarity(2)
    a = dist.sample(n, d, rng=np.random.RandomState(seed))
    b = dist.sample(n*d, rng=np.random.RandomState(seed))
    assert np.allclose(a.flatten(), b)


@pytest.mark.parametrize("d,p", [(3, 0), (5, 0.4), (10, 0.7), (50, 1.0)])
def test_cosine_intercept(d, p, rng):
    """Tests CosineSimilarity inverse cdf for finding intercepts."""
    pytest.importorskip('scipy')  # betaincinv

    num_samples = 250

    exp_dist = dists.UniformHypersphere(surface=True)
    act_dist = dists.CosineSimilarity(d)

    dots = exp_dist.sample(num_samples, d, rng=rng)[:, 0]

    # Find the desired intercept so that dots >= c with probability p
    c = act_dist.ppf(1 - p)
    assert np.allclose(np.sum(dots >= c) / float(num_samples), p, atol=0.05)


def test_distorarrayparam():
    """DistOrArrayParams can be distributions or samples."""
    class Test(object):
        dp = dists.DistOrArrayParam('dp',
                                    default=None, sample_shape=['*', '*'])

    inst = Test()
    inst.dp = dists.UniformHypersphere()
    assert isinstance(inst.dp, dists.UniformHypersphere)
    inst.dp = np.array([[1], [2], [3]])
    assert np.all(inst.dp == np.array([[1], [2], [3]]))
    with pytest.raises(ValueError):
        inst.dp = 'a'
    # Sample must have correct dims
    with pytest.raises(ValueError):
        inst.dp = np.array([1])


def test_distorarrayparam_sample_shape():
    """sample_shape dictates the shape of the sample that can be set."""
    class Test(object):
        dp = dists.DistOrArrayParam(
            'dp', default=None, sample_shape=['d1', 10])
        d1 = 4

    inst = Test()
    # Distributions are still cool
    inst.dp = dists.UniformHypersphere()
    assert isinstance(inst.dp, dists.UniformHypersphere)
    # Must be shape (4, 10)
    inst.dp = np.ones((4, 10))
    assert np.all(inst.dp == np.ones((4, 10)))
    with pytest.raises(ValidationError):
        inst.dp = np.ones((10, 4))
    assert np.all(inst.dp == np.ones((4, 10)))


def test_frozen():
    """Test attributes inherited from FrozenObject"""
    a = dists.Uniform(-0.3, 0.6)
    b = dists.Uniform(-0.3, 0.6)
    c = dists.Uniform(-0.2, 0.6)

    assert hash(a) == hash(a)
    assert hash(b) == hash(b)
    assert hash(c) == hash(c)

    assert a == b
    assert hash(a) == hash(b)
    assert a != c
    assert hash(a) != hash(c)  # not guaranteed, but highly likely
    assert b != c
    assert hash(b) != hash(c)  # not guaranteed, but highly likely



import warnings

import numpy as np

from scipy.linalg import inv

from nengolib.signal.lyapunov import balanced_transformation
from nengolib.signal.system import sys2zpk, LinearSystem

__all__ = ['pole_zero_cancel', 'modred', 'balance', 'balred']


def pole_zero_cancel(sys, tol=1e-8):
    """Pole/zero cancellation within a given tolerance.

    Sometimes referred to as the minimal realization in state-space. [#]_
    This (greedily) finds pole-zero pairs within a given tolerance, and
    removes them from the transfer function representation.

    Parameters
    ----------
    sys : :data:`linear_system_like`
       Linear system representation.
    tol : ``float``, optional
       Absolute tolerance to identify pole-zero pairs. Defaults to ``1e-8``.

    Returns
    -------
    reduced_sys : :class:`.LinearSystem`
       Reduced linear system in zero-pole-gain form.

    References
    ----------
    .. [#] http://www.mathworks.com/help/control/ref/minreal.html

    Examples
    --------
    See :doc:`notebooks/research/linear_model_reduction` for a notebook
    example.

    >>> from nengolib.signal import pole_zero_cancel, s
    >>> sys = (s - 1) / ((s - 1) * (s + 1))
    >>> assert pole_zero_cancel(sys) == 1 / (s + 1)
    """

    z, p, k = sys2zpk(sys)
    mz = np.ones(len(z), dtype=bool)  # start with all zeros
    mp = np.zeros(len(p), dtype=bool)  # and no poles
    for i, pole in enumerate(p):
        # search among the remaining zeros
        bad = np.where((np.abs(pole - z) <= tol) & mz)[0]
        if len(bad):  # cancel this pole with one of the zeros
            mz[bad[0]] = False
        else:  # include this pole
            mp[i] = True
    return LinearSystem((z[mz], p[mp], k), analog=sys.analog)


def modred(sys, keep_states, method='del'):
    """Reduces model order by eliminating a subset of states.

    Parameters
    ----------
    sys : :data:`linear_system_like`
       Linear system representation.
    keep_states : ``array_like``
       Subset of dimensions (integer indices between ``0`` and
       ``len(sys)-1``, inclusive) to keep.
    method : ``string``, optional
       Defaults to ``'del'``. Must be one of:

       * ``'del'`` : Delete the states entirely.

       * ``'dc'`` : Transform the remaining states to maintain the same
         DC gain. [#]_

    Returns
    -------
    reduced_sys : :class:`.LinearSystem`
       Reduced linear system in state-space form.

    See Also
    --------
    :func:`.balred`

    References
    ----------
    .. [#] http://www.mathworks.com/help/control/ref/modred.html

    Examples
    --------
    See :doc:`notebooks/research/linear_model_reduction` for a notebook
    example.
    """

    sys = LinearSystem(sys)
    A, B, C, D = sys.ss
    if not sys.analog:
        raise NotImplementedError("model reduction of digital filters not "
                                  "supported")

    mask = np.zeros(len(A), dtype=bool)
    mask[np.asarray(keep_states)] = True

    grm = np.where(mask)[0]
    brm = np.where(~mask)[0]
    glm = grm[:, None]
    blm = brm[:, None]

    A11 = A[glm, grm]
    A12 = A[glm, brm]
    A21 = A[blm, grm]
    A22 = A[blm, brm]
    B1 = B[mask, :]
    B2 = B[~mask, :]
    C1 = C[:, mask]
    C2 = C[:, ~mask]

    if method == 'del':
        RA = A11
        RB = B1
        RC = C1
        RD = D

    elif method == 'dc':
        A22I = inv(A22)
        RA = A11 - np.dot(A12, np.dot(A22I, A21))
        RB = B1 - np.dot(A12, np.dot(A22I, B2))
        RC = C1 - np.dot(C2, np.dot(A22I, A21))
        RD = D - np.dot(C2, np.dot(A22I, B2))
        # TODO: for discrete case, simply replace (-A22I) with inv(I - A22)

    else:
        raise ValueError("invalid method: '%s'" % (method,))

    return LinearSystem((RA, RB, RC, RD), analog=sys.analog)


def balance(sys):
    """Transforms a linear system to its balanced realization.

    Parameters
    ----------
    sys : :data:`linear_system_like`
       Linear system representation.

    Returns
    -------
    balanced_sys : :class:`.LinearSystem`
       Balanced linear system in state-space form.

    See Also
    --------
    :func:`.balred`
    :func:`.balanced_transformation`
    :class:`.Balanced`

    References
    ----------
    .. [#] https://www.mathworks.com/help/control/ref/balreal.html

    Examples
    --------
    See :doc:`notebooks/research/linear_model_reduction` for a notebook
    example.

    >>> from nengolib.signal import balance, s
    >>> before = 10 / ((s + 10) * (s + 20) * (s + 30) * (s + 40))
    >>> after = balance(before)

    Effect of balancing some arbitrary system:

    >>> import matplotlib.pyplot as plt
    >>> length = 500
    >>> plt.subplot(211)
    >>> plt.title("Impulse - Before")
    >>> plt.plot(before.ntrange(length), before.X.impulse(length))
    >>> plt.subplot(212)
    >>> plt.title("Impulse - After")
    >>> plt.plot(after.ntrange(length), after.X.impulse(length))
    >>> plt.xlabel("Time (s)")
    >>> plt.show()
    """

    sys = LinearSystem(sys)
    T, Tinv, _ = balanced_transformation(sys)
    return sys.transform(T, Tinv=Tinv)


def balred(sys, order, method='del'):
    """Reduces a linear system to given order using balance and modred.

    Parameters
    ----------
    sys : :data:`linear_system_like`
       Linear system representation.
    order : ``integer``
       Number of dimensions to keep.
    method : ``string``, optional
       Model order reduction method passed to :func:`.modred`.

    Returns
    -------
    reduced_sys : :class:`.LinearSystem`
       Balanced and reduced linear system in state-space form.

    See Also
    --------
    :func:`.balance`
    :func:`.modred`

    References
    ----------
    .. [#] https://www.mathworks.com/help/control/ref/balred.html

    Examples
    --------
    See :doc:`notebooks/research/linear_model_reduction` for a notebook
    example.
    """

    sys = LinearSystem(sys)
    if order < 1:
        raise ValueError("Invalid order (%s), must be at least 1" % (order,))
    if order >= len(sys):
        warnings.warn("Model is already of given order")
        return sys
    sys = balance(sys)
    keep_states = np.arange(order)  # keep largest eigenvalues
    return modred(sys, keep_states, method)



import numpy as np

from nengo.exceptions import ValidationError
from nengo.utils.compat import is_integer, is_number, range


class SemanticPointer(object):
    """A Semantic Pointer, based on Holographic Reduced Representations.

    Operators are overloaded so that ``+`` and ``-`` are addition,
    ``*`` is circular convolution, and ``~`` is the inversion operator.
    """

    def __init__(self, data, rng=None):
        if is_integer(data):
            if data < 1:
                raise ValidationError("Number of dimensions must be a "
                                      "positive int", attr='data', obj=self)

            self.randomize(data, rng=rng)
        else:
            try:
                len(data)
            except Exception:
                raise ValidationError(
                    "Must specify either the data or the length for a "
                    "SemanticPointer.", attr='data', obj=self)
            self.v = np.array(data, dtype=float)
            if len(self.v.shape) != 1:
                raise ValidationError("'data' must be a vector", 'data', self)

    def length(self):
        """Return the L2 norm of the vector."""
        return np.linalg.norm(self.v)

    def normalize(self):
        """Modify the vector to have an L2 norm of 1."""
        nrm = np.linalg.norm(self.v)
        if nrm > 0:
            self.v /= nrm

    def __str__(self):
        return str(self.v)

    def randomize(self, N=None, rng=None):
        """Set the vector to be a random vector with L2 norm of 1."""
        if N is None:
            N = len(self.v)
        if rng is None:
            rng = np.random
        self.v = rng.randn(N)
        self.normalize()

    def make_unitary(self):
        """Make the vector unitary."""
        fft_val = np.fft.fft(self.v)
        fft_imag = fft_val.imag
        fft_real = fft_val.real
        fft_norms = np.sqrt(fft_imag ** 2 + fft_real ** 2)
        fft_unit = fft_val / fft_norms
        self.v = np.array((np.fft.ifft(fft_unit)).real)

    def __add__(self, other):
        return SemanticPointer(data=self.v + other.v)

    def __iadd__(self, other):
        self.v += other.v
        return self

    def __neg__(self):
        return SemanticPointer(data=-self.v)

    def __sub__(self, other):
        return SemanticPointer(data=self.v - other.v)

    def __isub__(self, other):
        self.v -= other.v
        return self

    def __mul__(self, other):
        """Multiplication of two SemanticPointers is circular convolution.

        If multiplied by a scalar, we do normal multiplication.
        """
        if isinstance(other, SemanticPointer):
            return self.convolve(other)
        elif is_number(other):
            return SemanticPointer(data=self.v * other)
        else:
            raise NotImplementedError(
                "Can only multiply by SemanticPointers or scalars")

    def convolve(self, other):
        """Return the circular convolution of two SemanticPointers."""
        x = np.fft.ifft(np.fft.fft(self.v) * np.fft.fft(other.v)).real
        return SemanticPointer(data=x)

    def __rmul__(self, other):
        """Multiplication of two SemanticPointers is circular convolution.

        If mutliplied by a scaler, we do normal multiplication.
        """
        if isinstance(other, SemanticPointer):
            return self.convolve(other)
        elif is_number(other):
            return SemanticPointer(data=self.v * other)
        else:
            raise NotImplementedError(
                "Can only multiply by SemanticPointers or scalars")

    def __imul__(self, other):
        """Multiplication of two SemanticPointers is circular convolution.

        If mutliplied by a scaler, we do normal multiplication.
        """
        if isinstance(other, SemanticPointer):
            self.v = np.fft.ifft(np.fft.fft(self.v) *
                                 np.fft.fft(other.v)).real
        elif is_number(other):
            self.v *= other
        else:
            raise NotImplementedError(
                "Can only multiply by SemanticPointers or scalars")
        return self

    def compare(self, other):
        """Return the similarity between two SemanticPointers.

        This is the normalized dotproduct, or (equivalently), the cosine of
        the angle between the two vectors.
        """
        if isinstance(other, SemanticPointer):
            other = other.v
        scale = np.linalg.norm(self.v) * np.linalg.norm(other)
        if scale == 0:
            return 0
        return np.dot(self.v, other) / scale

    def dot(self, other):
        """Return the dot product of the two vectors."""
        if isinstance(other, SemanticPointer):
            other = other.v
        return np.dot(self.v, other)

    def distance(self, other):
        """Return a distance measure between the vectors.

        This is ``1-cos(angle)``, so that it is 0 when they are identical, and
        the distance gets larger as the vectors are farther apart.
        """
        return 1 - self.compare(other)

    def __invert__(self):
        """Return a reorganized vector that acts as an inverse for convolution.

        This reorganization turns circular convolution into circular
        correlation, meaning that ``A*B*~B`` is approximately ``A``.

        For the vector ``[1, 2, 3, 4, 5]``, the inverse is ``[1, 5, 4, 3, 2]``.
        """
        return SemanticPointer(data=self.v[-np.arange(len(self))])

    def __len__(self):
        """Return the number of dimensions in the vector."""
        return len(self.v)

    def copy(self):
        """Return another semantic pointer with the same data."""
        return SemanticPointer(data=self.v)

    def mse(self, other):
        """Return the mean-squared-error between two vectors."""
        return np.sum((self - other).v**2) / len(self.v)

    def get_convolution_matrix(self):
        """Return the matrix that does a circular convolution by this vector.

        This should be such that ``A*B == dot(A.get_convolution_matrix, B.v)``.
        """
        D = len(self.v)
        T = []
        for i in range(D):
            T.append([self.v[(i - j) % D] for j in range(D)])
        return np.array(T)



import numpy as np
import pytest

from nengo.utils.filter_design import expm, cont2discrete


def test_expm(rng):
    pytest.importorskip('scipy')
    import scipy.linalg as linalg
    for a in [np.eye(3), rng.randn(10, 10)]:
        assert np.allclose(linalg.expm(a), expm(a))


@pytest.mark.parametrize('dt', [1e-3, 1e-2, 1e-1])
def test_cont2discrete_zoh(dt):
    taus = np.logspace(-np.log10(dt) - 1, np.log10(dt) + 3, 30)

    # test with lowpass filter, using analytic solution
    for tau in taus:
        num, den = [1], [tau, 1]
        d = -np.expm1(-dt / tau)
        num0, den0 = [0, d], [1, d - 1]
        num1, den1, _ = cont2discrete((num, den), dt)
        assert np.allclose(num0, num1)
        assert np.allclose(den0, den1)

    # test with alpha filter, using analytic solution
    for tau in taus:
        num, den = [1], [tau**2, 2*tau, 1]
        a = dt / tau
        ea = np.exp(-a)
        num0 = [0, -a*ea + (1 - ea), ea*(a + ea - 1)]
        den0 = [1, -2 * ea, ea**2]
        num1, den1, _ = cont2discrete((num, den), dt)
        assert np.allclose(num0, num1)
        assert np.allclose(den0, den1)

    # test integrative filter, using analytic solution
    num, den = [1], [1, 0]
    num0, den0 = [0, dt], [1, -1]
    num1, den1, _ = cont2discrete((num, den), dt)
    assert np.allclose(num0, num1)
    assert np.allclose(den0, den1)



import numpy as np
from scipy.linalg import eig

from nengo.params import IntParam, NumberParam
from nengo.neurons import NeuronTypeParam
from nengo.synapses import SynapseParam

import nengo
from nengolib import Network
from nengolib.neurons import Tanh
from nengolib.networks.reservoir import Reservoir

__all__ = ['EchoState']


class EchoState(Network, Reservoir):
    """An Echo State Network (ESN) within a Nengo Reservoir.

    This creates a standard Echo State Network (ENS) as a Nengo network,
    defaulting to the standard set of assumptions of non-spiking Tanh units
    and a random recurrent weight matrix [1]_. This is based on the
    minimalist Python implementation from [2]_.

    The network takes some arbitrary time-varying vector as input, encodes it
    randomly, and filters it using nonlinear units and a random recurrent
    weight matrix normalized by its spectral radius.

    This class also inherits ``nengolib.networks.Reservoir``, and thus the
    optimal linear readout is solved for in the same way: the network is
    simulated on a test signal, and then a solver is used to optimize the
    decoding connection weights.

    References:
        [1] http://www.scholarpedia.org/article/Echo_state_network
        [2] http://minds.jacobs-university.de/mantas/code
    """

    n_neurons = IntParam('n_neurons', default=None, low=1)
    dimensions = IntParam('dimensions', default=None, low=1)
    dt = NumberParam('dt', low=0, low_open=True)
    recurrent_synapse = SynapseParam('recurrent_synapse')
    gain = NumberParam('gain', low=0, low_open=True)
    neuron_type = NeuronTypeParam('neuron_type')

    def __init__(self, n_neurons, dimensions, recurrent_synapse=0.005,
                 readout_synapse=None, radii=1.0, gain=1.25, rng=None,
                 neuron_type=Tanh(), include_bias=True, ens_seed=None,
                 label=None, seed=None, add_to_container=None, **ens_kwargs):
        """Initializes the Echo State Network.

        Parameters
        ----------
        n_neurons : int
            The number of neurons to use in the reservoir.
        dimensions : int
            The dimensionality of the input signal.
        recurrent_synapse : nengo.synapses.Synapse (Default: ``0.005``)
            Synapse used to filter the recurrent connection.
        readout_synapse : nengo.synapses.Synapse (Default: ``None``)
            Optional synapse to filter all of the outputs before solving
            for the linear readout. This is included in the connection to the
            ``output`` Node created within the network.
        radii : scalar or array_like, optional (Default: ``1``)
            The radius of each dimension of the input signal, used to normalize
            the incoming connection weights.
        gain : scalar, optional (Default: ``1.25``)
            A scalar gain on the recurrent connection weight matrix.
        rng : ``numpy.random.RandomState``, optional (Default: ``None``)
            Random state used to initialize all weights.
        neuron_type : ``nengo.neurons.NeuronType`` optional \
                      (Default: ``Tanh()``)
            Neuron model to use within the reservoir.
        include_bias : ``bool`` (Default: ``True``)
            Whether to include a bias current to the neural nonlinearity.
            This should be ``False`` if the neuron model already has a bias,
            e.g., ``LIF`` or ``LIFRate``.
        ens_seed : int, optional (Default: ``None``)
            Seed passed to the ensemble of neurons.
        """

        Network.__init__(self, label, seed, add_to_container)

        self.n_neurons = n_neurons
        self.dimensions = dimensions
        self.recurrent_synapse = recurrent_synapse
        self.radii = radii  # TODO: make array or scalar parameter?
        self.gain = gain
        self.rng = np.random if rng is None else rng
        self.neuron_type = neuron_type
        self.include_bias = include_bias

        self.W_in = (
            self.rng.rand(self.n_neurons, self.dimensions) - 0.5) / self.radii
        if self.include_bias:
            self.W_bias = self.rng.rand(self.n_neurons, 1) - 0.5
        else:
            self.W_bias = np.zeros((self.n_neurons, 1))
        self.W = self.rng.rand(self.n_neurons, self.n_neurons) - 0.5
        self.W *= self.gain / max(abs(eig(self.W)[0]))

        with self:
            self.ensemble = nengo.Ensemble(
                self.n_neurons, 1, neuron_type=self.neuron_type, seed=ens_seed,
                **ens_kwargs)
            self.input = nengo.Node(size_in=self.dimensions)

            pool = self.ensemble.neurons
            nengo.Connection(
                self.input, pool, transform=self.W_in, synapse=None)
            nengo.Connection(  # note the bias will be active during training
                nengo.Node(output=1, label="bias"), pool,
                transform=self.W_bias, synapse=None)
            nengo.Connection(
                self.ensemble.neurons, pool, transform=self.W,
                synapse=self.recurrent_synapse)

        Reservoir.__init__(
            self, self.input, pool, readout_synapse=readout_synapse,
            network=self)



import warnings

import nengo


def Integrator(recurrent_tau, n_neurons, dimensions, net=None, **kwargs):
    """An ensemble that accumulates input and maintains state.

    This is accomplished through scaling the input signal and recurrently
    connecting an ensemble to itself to maintain state.

    Parameters
    ----------
    recurrent_tau : float
        Time constant on the recurrent connection.
    n_neurons : int
        Number of neurons in the recurrently connected ensemble.
    dimensions : int
        Dimensionality of the input signal and ensemble.
    kwargs
        Keyword arguments passed through to ``nengo.Network``.

    Returns
    -------
    net : Network
        The newly built product network, or the provided ``net``.

    Attributes
    ----------
    net.ensemble : Ensemble
        The recurrently connected ensemble.
    net.input : Node
        Provides the input signal.
    """
    if net is None:
        kwargs.setdefault('label', "Integrator")
        net = nengo.Network(**kwargs)
    else:
        warnings.warn("The 'net' argument is deprecated.", DeprecationWarning)

    with net:
        net.input = nengo.Node(size_in=dimensions)
        net.ensemble = nengo.Ensemble(n_neurons, dimensions=dimensions)
        nengo.Connection(net.ensemble, net.ensemble, synapse=recurrent_tau)
        nengo.Connection(net.input, net.ensemble,
                         transform=recurrent_tau, synapse=None)
    net.output = net.ensemble
    return net



"""Implementation of the Nengo cache object (NCO) protocol.

Nengo cache objects store a Numpy array and some associated, picklable Python
object in a single, uncompressed file. These files are not platform independent
as they are optimized for fast reading and writing, and cached data is not
supposed to be shared across platforms.

The protocol version 0 is as follows:

* A header consisting of:
    * 3 bytes with the magic string 'NCO'
    * 1 unsigned byte indicating the protocol version
    * unsigned long int denoting the start of the Python object data
    * unsigned long int denoting the end of the Python object data
    * unsigned long int denoting the start of the array data
    * unsigned long int denoting the end of the array data
* Potentially some padding bytes.
* The Python object data pickled by the (c)pickle module using the highest
  available protocol.
* Potentially some padding bytes.
* The array data in NPY format.

Files will be written with padding to have both the Python object data and the
array data an alignment of 16 bytes.

The Numpy NPY format is documented here:
https://github.com/numpy/numpy/blob/master/doc/neps/npy-format.rst

As of legacy version 1 of the cache, multiple NCO files will be concatenated
into one file. The start and end of each subfile will be stored in a cache
index, but can also be recovered from reading the headers of the NCO files in
order as each one gives the start of the next header (corresponding to the
end of the array data).
"""

from __future__ import absolute_import

import os
import struct

import numpy as np

from .compat import ensure_bytes, pickle
from .cache import byte_align
from ..exceptions import CacheIOError


class Subfile(object):
    """A file-like object for limiting reads to a subrange of a file.

    This class only supports reading and seeking. Writing is not supported.

    Parameters
    ----------
    fileobj : file-like object
        Complete files.
    start : int
        Offset of the first readable position in the file.
    end : int
        Offset of the last readable position + 1 in the file.
    """
    def __init__(self, fileobj, start, end):
        self.fileobj = fileobj
        self.start = start
        self.end = end
        self.max_size = end - start

        self.fileobj.seek(start)

    def read(self, size=None):
        size = min(size, self.max_size) if size is not None else self.max_size
        self.max_size -= size
        return self.fileobj.read(size)

    def readline(self, size=None):
        size = min(size, self.max_size) if size is not None else self.max_size
        data = self.fileobj.readline(size)
        self.max_size = self.end - self.fileobj.tell()
        return data

    def seek(self, offset, whence=os.SEEK_SET):
        if whence == os.SEEK_CUR:
            offset = self.fileobj.tell() + offset
        elif whence == os.SEEK_SET:
            offset = self.start + offset
        elif whence == os.SEEK_END:
            offset = self.end + offset
        else:
            raise NotImplementedError()
        offset = max(self.start, min(self.end, offset))
        self.max_size = self.end - offset
        self.fileobj.seek(offset)

    def tell(self):
        return self.fileobj.tell() - self.start


MAGIC_STRING = ensure_bytes('NCO')
SUPPORTED_PROTOCOLS = [0]
HEADER_FORMAT = '@{}sBLLLL'.format(len(MAGIC_STRING))
HEADER_SIZE = struct.calcsize(HEADER_FORMAT)
ALIGNMENT = 16


def write(fileobj, metadata, array):
    """Writes a Nengo cache object.

    Parameters
    ----------
    fileobj : file-like object
        File object to write the data to.
    metadata : object
        Python object with metadata (will be pickled).
    array : ndarray
        Numpy array with the actual data to store.
    """
    start = fileobj.tell()
    pickle_start = byte_align(start + HEADER_SIZE, ALIGNMENT)
    fileobj.seek(pickle_start)
    pickle.dump(metadata, fileobj, pickle.HIGHEST_PROTOCOL)
    pickle_end = fileobj.tell()

    array_start = byte_align(pickle_end, ALIGNMENT)
    fileobj.seek(array_start)
    np.save(fileobj, array)
    array_end = fileobj.tell()

    header = struct.pack(
        HEADER_FORMAT, MAGIC_STRING, 0, pickle_start, pickle_end,
        array_start, array_end)
    fileobj.seek(start)
    fileobj.write(header)
    fileobj.seek(array_end)


def read(fileobj):
    """Reads a Nengo cache object.

    Parameters
    ----------
    fileobj : file-like object
        The file object to read from.

    Returns
    -------
    metadata, array
        Returns a tuple with the Python object containing the metadata as first
        element and the array with the actual data as second element.
    """
    header = fileobj.read(HEADER_SIZE)
    magic, version, pickle_start, pickle_end, array_start, array_end = (
        struct.unpack(HEADER_FORMAT, header))

    if magic != MAGIC_STRING:
        raise CacheIOError("Not a Nengo cache object file.")
    if version not in SUPPORTED_PROTOCOLS:
        raise CacheIOError(
            "NCO protocol version {} is not supported.".format(version))

    metadata = pickle.load(Subfile(fileobj, pickle_start, pickle_end))
    array = np.load(Subfile(fileobj, array_start, array_end))
    return metadata, array



import numpy as np
import pytest

import nengo
from nengo.utils.compat import range
from nengo.utils.functions import HilbertCurve
from nengo.utils.numpy import rmse, maxint


def test_sine_waves(Simulator, plt, seed):
    radius = 2
    dim = 5
    product = nengo.networks.Product(200, dim, radius, seed=seed)

    func_a = lambda t: np.sqrt(radius)*np.sin(np.arange(1, dim+1)*2*np.pi*t)
    func_b = lambda t: np.sqrt(radius)*np.sin(np.arange(dim, 0, -1)*2*np.pi*t)
    with product:
        input_a = nengo.Node(func_a)
        input_b = nengo.Node(func_b)
        nengo.Connection(input_a, product.input_a)
        nengo.Connection(input_b, product.input_b)
        p = nengo.Probe(product.output, synapse=0.005)

    with Simulator(product) as sim:
        sim.run(1.0)

    t = sim.trange()
    ideal = np.asarray(list(map(func_a, t))) * np.asarray(list(map(func_b, t)))
    delay = 0.013
    offset = np.where(t >= delay)[0]

    for i in range(dim):
        plt.subplot(dim+1, 1, i+1)
        plt.plot(t + delay, ideal[:, i])
        plt.plot(t, sim.data[p][:, i])
        plt.xlim(right=t[-1])
        plt.yticks((-2, 0, 2))

    assert rmse(ideal[:len(offset), :], sim.data[p][offset, :]) < 0.2


def test_direct_mode_with_single_neuron(Simulator, plt, seed):
    radius = 2
    dim = 5

    config = nengo.Config(nengo.Ensemble)
    config[nengo.Ensemble].neuron_type = nengo.Direct()
    with config:
        product = nengo.networks.Product(1, dim, radius, seed=seed)

    func_a = lambda t: np.sqrt(radius)*np.sin(np.arange(1, dim+1)*2*np.pi*t)
    func_b = lambda t: np.sqrt(radius)*np.sin(np.arange(dim, 0, -1)*2*np.pi*t)
    with product:
        input_a = nengo.Node(func_a)
        input_b = nengo.Node(func_b)
        nengo.Connection(input_a, product.input_a)
        nengo.Connection(input_b, product.input_b)
        p = nengo.Probe(product.output, synapse=0.005)

    with Simulator(product) as sim:
        sim.run(1.0)

    t = sim.trange()
    ideal = np.asarray(list(map(func_a, t))) * np.asarray(list(map(func_b, t)))
    delay = 0.013
    offset = np.where(t >= delay)[0]

    for i in range(dim):
        plt.subplot(dim+1, 1, i+1)
        plt.plot(t + delay, ideal[:, i])
        plt.plot(t, sim.data[p][:, i])
        plt.xlim(right=t[-1])
        plt.yticks((-2, 0, 2))

    assert rmse(ideal[:len(offset), :], sim.data[p][offset, :]) < 0.2


@pytest.mark.benchmark
@pytest.mark.slow
def test_product_benchmark(Simulator, analytics, rng):
    n_trials = 50
    hc = HilbertCurve(n=4)  # Increase n to cover the input space more densely
    duration = 5.           # Simulation duration (s)
    # Duration (s) to wait at the beginning to have a stable representation
    wait_duration = 0.5
    n_neurons = 100
    n_eval_points = 1000

    def stimulus_fn(t):
        return np.squeeze(hc(t / duration).T * 2 - 1)

    def run_trial():
        model = nengo.Network(seed=rng.randint(maxint))
        with model:
            model.config[nengo.Ensemble].n_eval_points = n_eval_points

            stimulus = nengo.Node(
                output=lambda t: stimulus_fn(max(0., t - wait_duration)),
                size_out=2)

            product_net = nengo.networks.Product(n_neurons, 1)
            nengo.Connection(stimulus[0], product_net.input_a)
            nengo.Connection(stimulus[1], product_net.input_b)
            probe_test = nengo.Probe(product_net.output)

            ens_direct = nengo.Ensemble(
                1, dimensions=2, neuron_type=nengo.Direct())
            result_direct = nengo.Node(size_in=1)
            nengo.Connection(stimulus, ens_direct)
            nengo.Connection(
                ens_direct, result_direct, function=lambda x: x[0] * x[1],
                synapse=None)
            probe_direct = nengo.Probe(result_direct)

        with Simulator(model) as sim:
            sim.run(duration + wait_duration, progress_bar=False)

        selection = sim.trange() > wait_duration
        test = sim.data[probe_test][selection]
        direct = sim.data[probe_direct][selection]
        return rmse(test, direct)

    error_data = [run_trial() for i in range(n_trials)]
    analytics.add_data(
        'error', error_data, "Multiplication RMSE. Shape: n_trials")


@pytest.mark.compare
def test_compare_product_benchmark(analytics_data, logger):
    stats = pytest.importorskip('scipy.stats')
    data1, data2 = (d['error'] for d in analytics_data)
    improvement = np.mean(data1) - np.mean(data2)
    p = np.ceil(1000. * 2. * stats.mannwhitneyu(
        data1, data2, alternative='two-sided')[1]) / 1000.
    logger.info("Multiplication improvement by %f (%.0f%%, p < %.3f)",
                improvement, (1. - np.mean(data2) / np.mean(data1)) * 100., p)
    assert improvement >= 0. or p >= 0.05



import numpy as np
from scipy import linalg
from scipy.signal import cont2discrete as _cont2discrete

from nengolib.signal.system import LinearSystem

__all__ = ['cont2discrete', 'discrete2cont']


def cont2discrete(sys, dt, method='zoh', alpha=None):
    """Convert linear system from continuous to discrete time-domain.

    This is a wrapper around :func:`scipy.signal.cont2discrete`, with the
    same interface (apart from the type of the first parameter).

    Parameters
    ----------
    sys : :data:`linear_system_like`
       Linear system representation.
    dt : ``float``
       Time-step for discrete simulation of target system.
    method : ``string``, optional
       Method of discretization. Defaults to zero-order hold discretization
       (``'zoh'``), which assumes that the input signal is held constant over
       each discrete time-step. [#]_
    alpha : ``float`` or ``None``, optional
       Weighting parameter for use with ``method='gbt'``.

    Returns
    -------
    discrete_sys : :class:`.LinearSystem`
       Discretized linear system (``analog=False``).

    See Also
    --------
    :func:`.discrete2cont`
    :func:`scipy.signal.cont2discrete`

    Notes
    -----
    Discretization is often performed automatically whenever needed;
    usually it is unnecessary to deal with this routine directly. One
    exception is when combining systems across domains (see example).

    References
    ----------
    .. [#] https://en.wikipedia.org/wiki/Discretization

    Examples
    --------
    Simulating an alpha synapse with a pure transmission delay:

    >>> from nengolib.signal import z, cont2discrete
    >>> from nengolib import Alpha
    >>> sys = Alpha(0.003)
    >>> dsys = z**(-20) * cont2discrete(sys, dt=sys.default_dt)
    >>> y = dsys.impulse(50)

    >>> assert np.allclose(np.sum(y), 1, atol=1e-3)
    >>> t = dsys.ntrange(len(y))

    >>> import matplotlib.pyplot as plt
    >>> plt.step(t, y, where='post')
    >>> plt.fill_between(t, np.zeros_like(y), y, step='post', alpha=.3)
    >>> plt.xlabel("Time (s)")
    >>> plt.show()
    """

    sys = LinearSystem(sys)
    if not sys.analog:
        raise ValueError("system (%s) is already discrete" % sys)
    return LinearSystem(
        _cont2discrete(sys.ss, dt=dt, method=method, alpha=alpha)[:-1],
        analog=False)


def discrete2cont(sys, dt, method='zoh', alpha=None):
    """Convert linear system from discrete to continuous time-domain.

    This is the inverse of :func:`.cont2discrete`. This will not work in
    general, for instance with the ZOH method when the system has discrete
    poles at ``0`` (e.g., systems with pure time-delay elements).

    Parameters
    ----------
    sys : :data:`linear_system_like`
       Linear system representation.
    dt : ``float``
       Time-step used to *undiscretize* ``sys``.
    method : ``string``, optional
       Method of discretization. Defaults to zero-order hold discretization
       (``'zoh'``), which assumes that the input signal is held constant over
       each discrete time-step.
    alpha : ``float`` or ``None``, optional
       Weighting parameter for use with ``method='gbt'``.

    Returns
    -------
    continuous_sys : :class:`.LinearSystem`
       Continuous linear system (``analog=True``).

    See Also
    --------
    :func:`.cont2discrete`
    :func:`scipy.signal.cont2discrete`

    Examples
    --------
    Converting a double-exponential synapse back and forth between domains:

    >>> from nengolib.signal import discrete2cont, cont2discrete
    >>> from nengolib import DoubleExp
    >>> sys = DoubleExp(0.005, 0.2)
    >>> assert dsys == discrete2cont(cont2discrete(sys, dt=0.1), dt=0.1)
    """

    sys = LinearSystem(sys)
    if sys.analog:
        raise ValueError("system (%s) is already continuous" % sys)

    if dt <= 0:
        raise ValueError("dt (%s) must be positive" % (dt,))

    ad, bd, cd, dd = sys.ss
    n = ad.shape[0]
    m = n + bd.shape[1]

    if method == 'gbt':
        if alpha is None or alpha < 0 or alpha > 1:
            raise ValueError("alpha (%s) must be in range [0, 1]" % (alpha,))

        I = np.eye(n)
        ar = linalg.solve(alpha*dt*ad.T + (1-alpha)*dt*I, ad.T - I).T
        M = I - alpha*dt*ar

        br = np.dot(M, bd) / dt
        cr = np.dot(cd, M)
        dr = dd - alpha*np.dot(cr, bd)

    elif method in ('bilinear', 'tustin'):
        return discrete2cont(sys, dt, method='gbt', alpha=0.5)

    elif method in ('euler', 'forward_diff'):
        return discrete2cont(sys, dt, method='gbt', alpha=0.0)

    elif method == 'backward_diff':
        return discrete2cont(sys, dt, method='gbt', alpha=1.0)

    elif method == 'zoh':
        M = np.zeros((m, m))
        M[:n, :n] = ad
        M[:n, n:] = bd
        M[n:, n:] = np.eye(bd.shape[1])
        E = linalg.logm(M) / dt

        ar = E[:n, :n]
        br = E[:n, n:]
        cr = cd
        dr = dd

    else:
        raise ValueError("invalid method: '%s'" % (method,))

    return LinearSystem((ar, br, cr, dr), analog=True)



from __future__ import absolute_import

import warnings
from distutils.version import LooseVersion

import numpy as np
import matplotlib
import matplotlib.pyplot as plt

from nengo.utils.compat import range
from nengo.utils.ensemble import tuning_curves

has_prop_cycle = LooseVersion(matplotlib.__version__) >= '1.5.0'

if has_prop_cycle:
    from cycler import cycler  # Dependency of MPL form 1.5.0 onward


def get_color_cycle():
    if has_prop_cycle:
        cycle = matplotlib.rcParams['axes.prop_cycle']
        # Apparently the 'color' key may not exist, so have to fail gracefully
        try:
            return [prop['color'] for prop in cycle]
        except KeyError:
            pass  # Fall back on deprecated axes.color_cycle
    return matplotlib.rcParams['axes.color_cycle']


def set_color_cycle(colors, ax=None):
    if has_prop_cycle:
        if ax is None:
            plt.rc('axes', prop_cycle=cycler('color', colors))
        else:
            ax.set_prop_cycle('color', colors)
    else:
        if ax is None:
            plt.rc('axes', color_cycle=colors)
        else:
            ax.set_color_cycle(colors)


def axis_size(ax=None):
    """Get axis width and height in pixels.

    Based on a StackOverflow response:
    https://stackoverflow.com/questions/19306510/
        determine-matplotlib-axis-size-in-pixels

    Parameters
    ----------
    ax : axis object
        The axes to determine the size of. Defaults to current axes.

    Returns
    -------
    width : float
        Width of axes in pixels.
    height : float
        Height of axes in pixels.
    """
    ax = plt.gca() if ax is None else ax
    fig = ax.figure
    bbox = ax.get_window_extent().transformed(fig.dpi_scale_trans.inverted())
    return bbox.width * fig.dpi, bbox.height * fig.dpi


def implot(plt, x, y, Z, ax=None, colorbar=True, **kwargs):
    """Image plot of general data (like imshow but with non-pixel axes).

    Parameters
    ----------
    plt : plot object
        Plot object, typically `matplotlib.pyplot`.
    x : (M,) array_like
        Vector of x-axis points, must be linear (equally spaced).
    y : (N,) array_like
        Vector of y-axis points, must be linear (equally spaced).
    Z : (M, N) array_like
        Matrix of data to be displayed, the value at each (x, y) point.
    ax : axis object (optional)
        A specific axis to plot on (defaults to `plt.gca()`).
    colorbar: boolean (optional)
        Whether to plot a colorbar.
    **kwargs
        Additional arguments for `ax.imshow`.
    """
    ax = plt.gca() if ax is None else ax

    def is_linear(x):
        diff = np.diff(x)
        return np.allclose(diff, diff[0])

    assert is_linear(x) and is_linear(y)
    image = ax.imshow(Z, aspect='auto', extent=(x[0], x[-1], y[-1], y[0]),
                      **kwargs)
    if colorbar:
        plt.colorbar(image, ax=ax)


def rasterplot(time, spikes, ax=None, use_eventplot=False, **kwargs):  # noqa
    """Generate a raster plot of the provided spike data.

    Parameters
    ----------
    time : array
        Time data from the simulation
    spikes : array
        The spike data with columns for each neuron and 1s indicating spikes
    ax : matplotlib.axes.Axes, optional (Default: None)
        The figure axes to plot into. If None, we will use current axes.
    use_eventplot : boolean, optional (Default: False)
        Whether to use the new Matplotlib `eventplot` routine. It is slower
        and makes larger image files, so we do not use it by default.

    Returns
    -------
    ax : matplotlib.axes.Axes
        The axes that were plotted into

    Examples
    --------
    >>> import nengo
    >>> with nengo.Network() as net:
    ...     a = nengo.Ensemble(20, 1)
    ...     p = nengo.Probe(a.neurons)
    >>> with nengo.Simulator(net) as sim:
    ...     sim.run(1)
    >>> rasterplot(sim.trange(), sim.data[p])
    """
    n_times, n_neurons = spikes.shape

    if ax is None:
        ax = plt.gca()

    if use_eventplot and not hasattr(ax, 'eventplot'):
        warnings.warn("Matplotlib version %s does not have 'eventplot'. "
                      "Falling back to non-eventplot version."
                      % matplotlib.__version__)
        use_eventplot = False

    colors = kwargs.pop('colors', None)
    if colors is None:
        color_cycle = get_color_cycle()
        colors = [color_cycle[i % len(color_cycle)] for i in range(n_neurons)]

    # --- plotting
    if use_eventplot:
        spiketimes = [time[s > 0].ravel() for s in spikes.T]
        for ix in range(n_neurons):
            if spiketimes[ix].size == 0:
                spiketimes[ix] = np.array([-np.inf])

        # hack to make 'eventplot' count from 1 instead of 0
        spiketimes = [np.array([-np.inf])] + spiketimes
        colors = [['k']] + colors

        ax.eventplot(spiketimes, colors=colors, **kwargs)
    else:
        kwargs.setdefault('linestyle', 'None')
        kwargs.setdefault('marker', '|')
        # Default markersize determined by matching eventplot
        ax_height = axis_size(ax)[1]
        markersize = max(ax_height * 0.8 / n_neurons, 1)
        # For 1 - 3 neurons, we need an extra fudge factor to match eventplot
        markersize -= max(4 - n_neurons, 0) ** 2 * ax_height * 0.005
        kwargs.setdefault('markersize', markersize)
        kwargs.setdefault('markeredgewidth', 1)

        for i in range(n_neurons):
            spiketimes = time[spikes[:, i] > 0].ravel()
            ax.plot(spiketimes, np.zeros_like(spiketimes) + (i + 1),
                    color=colors[i], **kwargs)

    # --- set axes limits
    if n_times > 1:
        ax.set_xlim(time[0], time[-1])

    ax.set_ylim(n_neurons + 0.6, 0.4)
    if n_neurons < 5:
        # make sure only integer ticks for small neuron numbers
        ax.set_yticks(np.arange(1, n_neurons + 1))

    # --- remove ticks as these are distracting in rasters
    ax.xaxis.set_ticks_position('none')
    ax.yaxis.set_ticks_position('none')

    return ax


def plot_tuning_curves(ensemble, sim, connection=None, ax=None):
    """Plot tuning curves for the given ensemble and simulator.

    If a connection is provided, the decoders will be used to set
    the colours of the tuning curves.
    """

    if ax is None:
        ax = plt.gca()

    evals, t_curves = tuning_curves(ensemble, sim)

    if connection is not None:
        if connection.dimensions > 1:
            warnings.warn("Ignoring dimensions > 1 in plot_tuning_curves")
        cm = plt.cm.ScalarMappable(cmap=plt.cm.coolwarm)
        set_color_cycle(cm.to_rgba(sim.data[connection].decoders[0]), ax=ax)
    ax.plot(evals, t_curves)



import hashlib
import inspect
import importlib
import os
import re
from fnmatch import fnmatch
import warnings

import matplotlib
import numpy as np
import pytest

import nengo
import nengo.utils.numpy as npext
from nengo.neurons import (Direct, LIF, LIFRate, RectifiedLinear,
                           Sigmoid, SpikingRectifiedLinear)
from nengo.rc import rc
from nengo.utils.compat import ensure_bytes, is_string
from nengo.utils.testing import Analytics, Logger, Plotter


class TestConfig(object):
    """Parameters affecting all Nengo tests.

    These are essentially global variables used by py.test to modify aspects
    of the Nengo tests. We collect them in this class to provide a
    mini namespace and to avoid using the ``global`` keyword.

    The values below are defaults. The functions in the remainder of this
    module modify these values accordingly.
    """

    test_seed = 0  # changing this will change seeds for all tests
    Simulator = nengo.Simulator
    RefSimulator = nengo.Simulator
    neuron_types = [
        Direct, LIF, LIFRate, RectifiedLinear, Sigmoid, SpikingRectifiedLinear
    ]
    compare_requested = False

    @classmethod
    def is_sim_overridden(cls):
        return cls.Simulator is not nengo.Simulator

    @classmethod
    def is_refsim_overridden(cls):
        return cls.RefSimulator is not nengo.Simulator

    @classmethod
    def is_skipping_frontend_tests(cls):
        return cls.is_sim_overridden() or cls.is_refsim_overridden()


def pytest_configure(config):
    matplotlib.use('agg')
    warnings.simplefilter('always')

    if config.getoption('simulator'):
        TestConfig.Simulator = load_class(config.getoption('simulator')[0])
    if config.getoption('ref_simulator'):
        refsim = config.getoption('ref_simulator')[0]
        TestConfig.RefSimulator = load_class(refsim)

    if config.getoption('neurons'):
        ntypes = config.getoption('neurons')[0].split(',')
        TestConfig.neuron_types = [load_class(n) for n in ntypes]

    if config.getoption('seed_offset'):
        TestConfig.test_seed = config.getoption('seed_offset')[0]

    TestConfig.compare_requested = config.getvalue('compare') is not None


def load_class(fully_qualified_name):
    mod_name, cls_name = fully_qualified_name.rsplit('.', 1)
    mod = importlib.import_module(mod_name)
    return getattr(mod, cls_name)


@pytest.fixture(scope="session")
def Simulator(request):
    """The Simulator class being tested.

    Please use this, and not ``nengo.Simulator`` directly. If the test is
    reference simulator specific, then use ``RefSimulator`` below.
    """
    return TestConfig.Simulator


@pytest.fixture(scope="session")
def RefSimulator(request):
    """The reference simulator.

    Please use this if the test is reference simulator specific.
    Other simulators may choose to implement the same API as the
    reference simulator; this allows them to test easily.
    """
    return TestConfig.RefSimulator


def recorder_dirname(request, name):
    """Returns the directory to put test artifacts in.

    Test artifacts produced by Nengo include plots and analytics.

    Note that the return value might be None, which indicates that the
    artifacts should not be saved.
    """
    record = request.config.getvalue(name)
    if is_string(record):
        return record
    elif not record:
        return None

    simulator, nl = TestConfig.RefSimulator, None
    if 'Simulator' in request.funcargnames:
        simulator = request.getfixturevalue('Simulator')
    # 'nl' stands for the non-linearity used in the neuron equation
    if 'nl' in request.funcargnames:
        nl = request.getfixturevalue('nl')
    elif 'nl_nodirect' in request.funcargnames:
        nl = request.getfixturevalue('nl_nodirect')

    dirname = "%s.%s" % (simulator.__module__, name)
    if nl is not None:
        dirname = os.path.join(dirname, nl.__name__)
    return dirname


def parametrize_function_name(request, function_name):
    """Creates a unique name for a test function.

    The unique name accounts for values passed through
    ``pytest.mark.parametrize``.

    This function is used when naming plots saved through the ``plt`` fixture.
    """
    suffixes = []
    if 'parametrize' in request.keywords:
        argnames = request.keywords['parametrize'].args[::2]
        argnames = [x.strip() for names in argnames for x in names.split(',')]
        for name in argnames:
            value = request.getfixturevalue(name)
            if inspect.isclass(value):
                value = value.__name__
            suffixes.append('{}={}'.format(name, value))
    return '+'.join([function_name] + suffixes)


@pytest.fixture
def plt(request):
    """A pyplot-compatible plotting interface.

    Please use this if your test creates plots.

    This will keep saved plots organized in a simulator-specific folder,
    with an automatically generated name. savefig() and close() will
    automatically be called when the test function completes.

    If you need to override the default filename, set `plt.saveas` to
    the desired filename.
    """
    dirname = recorder_dirname(request, 'plots')
    plotter = Plotter(
        dirname, request.module.__name__,
        parametrize_function_name(request, request.function.__name__))
    request.addfinalizer(lambda: plotter.__exit__(None, None, None))
    return plotter.__enter__()


@pytest.fixture
def analytics(request):
    """An object to store data for analytics.

    Please use this if you're concerned that accuracy or speed may regress.

    This will keep saved data organized in a simulator-specific folder,
    with an automatically generated name. Raw data (for later processing)
    can be saved with ``analytics.add_raw_data``; these will be saved in
    separate compressed ``.npz`` files. Summary data can be saved with
    ``analytics.add_summary_data``; these will be saved
    in a single ``.csv`` file.
    """
    dirname = recorder_dirname(request, 'analytics')
    analytics = Analytics(
        dirname, request.module.__name__,
        parametrize_function_name(request, request.function.__name__))
    request.addfinalizer(lambda: analytics.__exit__(None, None, None))
    return analytics.__enter__()


@pytest.fixture
def analytics_data(request):
    paths = request.config.getvalue('compare')
    function_name = parametrize_function_name(request, re.sub(
        '^test_[a-zA-Z0-9]*_', 'test_', request.function.__name__, count=1))
    return [Analytics.load(
        p, request.module.__name__, function_name) for p in paths]


@pytest.fixture
def logger(request):
    """A logging.Logger object.

    Please use this if your test emits log messages.

    This will keep saved logs organized in a simulator-specific folder,
    with an automatically generated name.
    """
    dirname = recorder_dirname(request, 'logs')
    logger = Logger(
        dirname, request.module.__name__,
        parametrize_function_name(request, request.function.__name__))
    request.addfinalizer(lambda: logger.__exit__(None, None, None))
    return logger.__enter__()


def function_seed(function, mod=0):
    """Generates a unique seed for the given test function.

    The seed should be the same across all machines/platforms.
    """
    c = function.__code__

    # get function file path relative to Nengo directory root
    nengo_path = os.path.abspath(os.path.dirname(nengo.__file__))
    path = os.path.relpath(c.co_filename, start=nengo_path)

    # take start of md5 hash of function file and name, should be unique
    hash_list = os.path.normpath(path).split(os.path.sep) + [c.co_name]
    hash_string = ensure_bytes('/'.join(hash_list))
    i = int(hashlib.md5(hash_string).hexdigest()[:15], 16)
    s = (i + mod) % npext.maxint
    int_s = int(s)  # numpy 1.8.0 bug when RandomState on long type inputs
    assert type(int_s) == int  # should not still be a long because < maxint
    return int_s


def get_item_name(item):
    """Get a unique backend-independent name for an item (test function)."""
    item_abspath, item_name = str(item.fspath), item.location[2]
    nengo_path = os.path.abspath(os.path.dirname(nengo.__file__))
    item_relpath = os.path.relpath(item_abspath, start=nengo_path)
    item_relpath = os.path.join('nengo', item_relpath)
    item_relpath = item_relpath.replace(os.sep, '/')
    return '%s:%s' % (item_relpath, item_name)


@pytest.fixture
def rng(request):
    """A seeded random number generator.

    This should be used in lieu of np.random because we control its seed.
    """
    # add 1 to seed to be different from `seed` fixture
    seed = function_seed(request.function, mod=TestConfig.test_seed + 1)
    return np.random.RandomState(seed)


@pytest.fixture
def seed(request):
    """A seed for seeding Networks.

    This should be used in lieu of an integer seed so that we can ensure that
    tests are not dependent on specific seeds.
    """
    return function_seed(request.function, mod=TestConfig.test_seed)


def pytest_generate_tests(metafunc):
    marks = [
        getattr(pytest.mark, m.name)(*m.args, **m.kwargs)
        for m in getattr(metafunc.function, 'pytestmark', [])]

    def mark_nl(nl):
        if nl is Sigmoid:
            nl = pytest.param(nl, marks=[pytest.mark.filterwarnings(
                'ignore:overflow encountered in exp')] + marks)
        return nl

    if "nl" in metafunc.funcargnames:
        metafunc.parametrize(
            "nl", [mark_nl(nl) for nl in TestConfig.neuron_types])
    if "nl_nodirect" in metafunc.funcargnames:
        nodirect = [mark_nl(n) for n in TestConfig.neuron_types
                    if n is not Direct]
        metafunc.parametrize("nl_nodirect", nodirect)


def pytest_runtest_setup(item):  # noqa: C901
    rc.reload_rc([])
    rc.set('decoder_cache', 'enabled', 'False')
    rc.set('exceptions', 'simplified', 'False')

    if not hasattr(item, 'obj'):
        return  # Occurs for doctests, possibly other weird tests

    conf = item.config
    test_uses_compare = getattr(item.obj, 'compare', None) is not None
    test_uses_sim = 'Simulator' in item.fixturenames
    test_uses_refsim = 'RefSimulator' in item.fixturenames
    tests_frontend = not (test_uses_sim or test_uses_refsim)

    if getattr(item.obj, 'example', None) and not conf.getvalue('noexamples'):
        pytest.skip("examples not requested")
    elif getattr(item.obj, 'slow', None) and not conf.getvalue('slow'):
        pytest.skip("slow tests not requested")
    elif not TestConfig.compare_requested and test_uses_compare:
        pytest.skip("compare tests not requested")
    elif TestConfig.is_skipping_frontend_tests() and tests_frontend:
        pytest.skip("frontend tests not run for alternate backends")
    elif (TestConfig.is_skipping_frontend_tests()
          and test_uses_refsim
          and not TestConfig.is_refsim_overridden()):
        pytest.skip("RefSimulator not overridden")
    elif (TestConfig.is_skipping_frontend_tests()
          and test_uses_sim
          and not TestConfig.is_sim_overridden()):
        pytest.skip("Simulator not overridden")
    elif getattr(item.obj, 'noassertions', None):
        options = []
        for fixture, option in [('analytics', 'analytics'),
                                ('plt', 'plots'),
                                ('logger', 'logs')]:
            if fixture in item.fixturenames and not conf.getvalue(option):
                options.append(option)
        if len(options) > 0:
            pytest.skip("%s not requested" % " and ".join(options))

    if not tests_frontend:
        item_name = get_item_name(item)

        for test, reason in TestConfig.Simulator.unsupported:
            # We add a '*' before test to eliminate the surprise of needing
            # a '*' before the name of a test function.
            if fnmatch(item_name, '*' + test):
                pytest.xfail(reason)


def pytest_terminal_summary(terminalreporter):
    reports = terminalreporter.getreports('passed')
    if not reports or terminalreporter.config.getvalue('compare') is None:
        return
    terminalreporter.write_sep("=", "PASSED")
    for rep in reports:
        for name, content in rep.sections:
            terminalreporter.writer.sep("-", name)
            terminalreporter.writer.line(content)



import numpy as np
import pytest

from nengo.utils.numpy import rmse

from nengolib.signal.reduction import pole_zero_cancel, modred, balred
from nengolib import Lowpass, Alpha
from nengolib.signal import LinearSystem, balanced_transformation, shift
from nengolib.testing import warns


def test_minreal():
    sys1 = Lowpass(0.05)
    sys2 = Lowpass(0.01)
    sys3 = sys2*sys2
    sys4 = LinearSystem(2)

    assert pole_zero_cancel(sys1) == sys1
    assert pole_zero_cancel(sys3) == sys3
    assert pole_zero_cancel(sys3/sys2, tol=1e-4) == sys2  # numpy 1.9.2
    assert (pole_zero_cancel(sys1*sys2/sys3, tol=1e-4) ==
            sys1/sys2)  # numpy 1.9.2

    assert (pole_zero_cancel(sys2/sys3, tol=1e-4) ==
            pole_zero_cancel(~sys2))  # numpy 1.9.2
    assert (pole_zero_cancel(sys3*sys4) ==
            pole_zero_cancel(sys4)*pole_zero_cancel(sys3))


def test_modred(rng):
    dt = 0.001
    isys = Lowpass(0.05)
    noise = 0.5*Lowpass(0.01) + 0.5*Alpha(0.005)
    p = 0.999
    sys = p*isys + (1-p)*noise

    T, Tinv, S = balanced_transformation(sys)
    balsys = sys.transform(T, Tinv)

    # Keeping just the best state should remove the 3 noise dimensions
    # Discarding the lowest state should do at least as well
    for keep_states in (S.argmax(),
                        list(set(range(len(sys))) - set((S.argmin(),)))):
        delsys = modred(balsys, keep_states)
        assert delsys.order_den == np.asarray(keep_states).size

        u = rng.normal(size=2000)

        expected = sys.filt(u, dt)
        actual = delsys.filt(u, dt)
        assert rmse(expected, actual) < 1e-4

        step = np.zeros(2000)
        step[50:] = 1.0
        dcsys = modred(balsys, keep_states, method='dc')
        assert np.allclose(dcsys.dcgain, balsys.dcgain)

        # use of shift related to nengo issue #938
        assert not sys.has_passthrough
        assert dcsys.has_passthrough
        expected = shift(sys.filt(step, dt))
        actual = dcsys.filt(step, dt)
        assert rmse(expected, actual) < 1e-4


def test_invalid_modred():
    with pytest.raises(ValueError):
        modred(Lowpass(0.1), 0, method='zoh')


def test_balred(rng):
    dt = 0.001
    sys = Alpha(0.01) + Lowpass(0.001)

    u = rng.normal(size=2000)
    expected = sys.filt(u, dt)

    def check(order, within, tol, method='del'):
        redsys = balred(sys, order, method=method)
        assert redsys.order_den <= order
        actual = redsys.filt(u, dt)
        assert abs(rmse(expected, actual) - within) < tol

    with warns(UserWarning):
        check(4, 0, 1e-13)
    with warns(UserWarning):
        check(3, 0, 1e-13)
    check(2, 0.03, 0.01)
    check(1, 0.3, 0.1)


def test_invalid_balred():
    with pytest.raises(ValueError):
        balred(Lowpass(0.1), 0)



from __future__ import absolute_import

import inspect
import itertools
import logging
import os
import re
import sys
import threading
import time

import numpy as np

from .compat import is_string, reraise
from .logging import CaptureLogHandler, console_formatter


class Mock(object):
    def __init__(self, *args, **kwargs):
        pass

    def __call__(self, *args, **kwargs):
        return Mock()

    def __getitem__(self, key):
        return Mock()

    def __iter__(self):
        return iter([])

    def __mul__(self, other):
        return 1.0

    @classmethod
    def __getattr__(cls, name):
        if name in ('__file__', '__path__'):
            return '/dev/null'
        elif name[0] == name[0].upper():
            mockType = type(name, (), {})
            mockType.__module__ = __name__
            return mockType
        else:
            return Mock()


class Recorder(object):
    def __init__(self, dirname, module_name, function_name):
        self.dirname = dirname
        self.module_name = module_name
        self.function_name = function_name

        if self.record and not os.path.exists(self.dirname):
            os.makedirs(self.dirname)

    @property
    def record(self):
        return self.dirname is not None

    @property
    def dirname(self):
        return self._dirname

    @dirname.setter
    def dirname(self, _dirname):
        if _dirname is not None and not os.path.exists(_dirname):
            os.makedirs(_dirname)

        self._dirname = _dirname

    def get_filename(self, ext=''):
        modparts = self.module_name.split('.')[1:]
        if 'tests' in modparts:
            modparts.remove('tests')
        return "%s.%s.%s" % ('.'.join(modparts), self.function_name, ext)

    def get_filepath(self, ext=''):
        if not self.record:
            raise ValueError("Cannot construct path when not recording")
        return os.path.join(self.dirname, self.get_filename(ext=ext))

    def __enter__(self):
        raise NotImplementedError()

    def __exit__(self, type, value, traceback):
        raise NotImplementedError()


class Plotter(Recorder):
    def __enter__(self):
        if self.record:
            import matplotlib.pyplot as plt
            self.plt = plt
        else:
            self.plt = Mock()
        return self.plt

    def __exit__(self, type, value, traceback):
        if self.record:
            if hasattr(self.plt, 'saveas') and self.plt.saveas is None:
                del self.plt.saveas
                self.plt.close('all')
                return

            if hasattr(self.plt, 'saveas'):
                self.filename = self.plt.saveas
                del self.plt.saveas
            else:
                self.filename = self.get_filename(ext='pdf')

            if len(self.plt.gcf().get_axes()) > 0:
                # tight_layout errors if no axes are present
                self.plt.tight_layout()

            savefig_kw = {'bbox_inches': 'tight'}
            if hasattr(self.plt, 'bbox_extra_artists'):
                savefig_kw['bbox_extra_artists'] = self.plt.bbox_extra_artists
                del self.plt.bbox_extra_artists

            self.plt.savefig(os.path.join(self.dirname, self.filename),
                             **savefig_kw)
            self.plt.close('all')


class Analytics(Recorder):
    DOC_KEY = 'documentation'

    def __init__(self, dirname, module_name, function_name):
        super(Analytics, self).__init__(dirname, module_name, function_name)

        self.data = {}
        self.doc = {}

    @staticmethod
    def load(path, module, function_name):
        modparts = module.split('.')
        modparts = modparts[1:]
        modparts.remove('tests')

        return np.load(os.path.join(path, "%s.%s.npz" % (
            '.'.join(modparts), function_name)))

    def __enter__(self):
        return self

    def add_data(self, name, data, doc=""):
        if name == self.DOC_KEY:
            raise ValueError("The name '{}' is reserved.".format(self.DOC_KEY))

        if self.record:
            self.data[name] = data
            if doc != "":
                self.doc[name] = doc

    def save_data(self):
        if len(self.data) == 0:
            return

        npz_data = dict(self.data)
        if len(self.doc) > 0:
            npz_data.update({self.DOC_KEY: self.doc})
        np.savez(self.get_filepath(ext='npz'), **npz_data)

    def __exit__(self, type, value, traceback):
        if self.record:
            self.save_data()


class Logger(Recorder):
    def __enter__(self):
        if self.record:
            self.handler = CaptureLogHandler()
            self.handler.setFormatter(console_formatter)
            self.logger = logging.getLogger()
            self.logger.addHandler(self.handler)
            self.old_level = self.logger.getEffectiveLevel()
            self.logger.setLevel(logging.INFO)
            self.logger.info("=== Test run at %s ===",
                             time.strftime("%Y-%m-%d %H:%M:%S"))
        else:
            self.logger = Mock()
        return self.logger

    def __exit__(self, type, value, traceback):
        if self.record:
            self.logger.removeHandler(self.handler)
            self.logger.setLevel(self.old_level)
            with open(self.get_filepath(ext='txt'), 'a') as fp:
                fp.write(self.handler.stream.getvalue())
            self.handler.close()
            del self.handler


def allclose(t, targets, signals,  # noqa: C901
             atol=1e-8, rtol=1e-5, buf=0, delay=0,
             plt=None, show=False, labels=None, individual_results=False):
    """Ensure all signal elements are within tolerances.

    Allows for delay, removing the beginning of the signal, and plotting.

    Parameters
    ----------
    t : array_like (T,)
        Simulation time for the points in `target` and `signals`.
    targets : array_like (T, 1) or (T, N)
        Reference signal or signals for error comparison.
    signals : array_like (T, N)
        Signals to be tested against the target signals.
    atol, rtol : float
        Absolute and relative tolerances.
    buf : float
        Length of time (in seconds) to remove from the beginnings of signals.
    delay : float
        Amount of delay (in seconds) to account for when doing comparisons.
    plt : matplotlib.pyplot or mock
        Pyplot interface for plotting the results, unless it's mocked out.
    show : bool
        Whether to show the plot immediately.
    labels : list of string, length N
        Labels of each signal to use when plotting.
    individual_results : bool
        If True, returns a separate `allclose` result for each signal.
    """
    t = np.asarray(t)
    dt = t[1] - t[0]
    assert t.ndim == 1
    assert np.allclose(np.diff(t), dt)

    targets = np.asarray(targets)
    signals = np.asarray(signals)
    if targets.ndim == 1:
        targets = targets.reshape((-1, 1))
    if signals.ndim == 1:
        signals = signals.reshape((-1, 1))
    assert targets.ndim == 2 and signals.ndim == 2
    assert t.size == targets.shape[0]
    assert t.size == signals.shape[0]
    assert targets.shape[1] == 1 or targets.shape[1] == signals.shape[1]

    buf = int(np.round(buf / dt))
    delay = int(np.round(delay / dt))
    slice1 = slice(buf, len(t) - delay)
    slice2 = slice(buf + delay, None)

    if plt is not None:
        if labels is None:
            labels = [None] * len(signals)
        elif is_string(labels):
            labels = [labels]

        colors = ['b', 'g', 'r', 'c', 'm', 'y', 'k']

        def plot_target(ax, x, b=0, c='k'):
            bound = atol + rtol * np.abs(x)
            y = x - b
            ax.plot(t[slice2], y[slice1], c + ':')
            ax.plot(t[slice2], (y + bound)[slice1], c + '--')
            ax.plot(t[slice2], (y - bound)[slice1], c + '--')

        # signal plot
        ax = plt.subplot(2, 1, 1)
        for y, label in zip(signals.T, labels):
            ax.plot(t, y, label=label)

        if targets.shape[1] == 1:
            plot_target(ax, targets[:, 0], c='k')
        else:
            color_cycle = itertools.cycle(colors)
            for x in targets.T:
                plot_target(ax, x, c=next(color_cycle))

        ax.set_ylabel('signal')
        if labels[0] is not None:
            lgd = ax.legend(loc='upper left', bbox_to_anchor=(1., 1.))
            plt.bbox_extra_artists = (lgd,)

        ax = plt.subplot(2, 1, 2)
        if targets.shape[1] == 1:
            x = targets[:, 0]
            plot_target(ax, x, b=x, c='k')
            for y, label in zip(signals.T, labels):
                ax.plot(t[slice2], y[slice2] - x[slice1])
        else:
            color_cycle = itertools.cycle(colors)
            for x, y, label in zip(targets.T, signals.T, labels):
                c = next(color_cycle)
                plot_target(ax, x, b=x, c=c)
                ax.plot(t[slice2], y[slice2] - x[slice1], c, label=label)

        ax.set_xlabel('time')
        ax.set_ylabel('error')

        if show:
            plt.show()

    if individual_results:
        if targets.shape[1] == 1:
            return [np.allclose(y[slice2], targets[slice1, 0],
                                atol=atol, rtol=rtol) for y in signals.T]
        else:
            return [np.allclose(y[slice2], x[slice1], atol=atol, rtol=rtol)
                    for x, y in zip(targets.T, signals.T)]
    else:
        return np.allclose(signals[slice2, :], targets[slice1, :],
                           atol=atol, rtol=rtol)


def find_modules(root_path, prefix=None, pattern='^test_.*\\.py$'):
    """Find matching modules (files) in all subdirectories of a given path.

    Parameters
    ----------
    root_path : string
        The path of the directory in which to begin the search.
    prefix : string or list, optional
        A string or list of strings to append to each returned modules list.
    pattern : string, optional
        A regex pattern for matching individual file names. Defaults to
        looking for all testing scripts.

    Returns
    -------
    modules : list
        A list of modules. Each item in the list is a list of strings
        containing the module path.
    """
    prefix = [] if prefix is None else prefix
    if is_string(prefix):
        prefix = [prefix]
    elif not isinstance(prefix, list):
        raise TypeError("Invalid prefix type '%s'" % type(prefix).__name__)

    modules = []
    for path, dirs, files in os.walk(root_path):
        base = prefix + os.path.relpath(path, root_path).split(os.sep)
        for filename in files:
            if re.search(pattern, filename):
                name, ext = os.path.splitext(filename)
                modules.append(base + [name])

    return modules


def load_functions(modules, pattern='^test_', arg_pattern='^Simulator$'):
    """Load matching functions from a list of modules.

    Parameters
    ----------
    modules : list
        A list of testing modules to load, generated by `find_testmodules`.
    pattern : string, optional
        A regex pattern for matching the function names. Defaults to looking
        for all testing functions.
    arg_pattern : string, optional
        A regex pattern for matching the argument names. At least one argument
        must match the pattern. Defaults to selecting tests with
        a 'Simulator' argument.

    Returns
    -------
    tests : dict
        A dictionary of test functions, where the key is composed of the
        module and function name, and the value is the function handle.

    Examples
    --------
    To load all Nengo tests and add them to the current namespace, do

        nengo_dir = os.path.dirname(nengo.__file__)
        modules = find_modules(nengo_dir, prefix='nengo')
        tests = load_functions(modules)
        locals().update(tests)

    Notes
    -----
    - This was created to load py.test tests. Therefore, this function also
      loads functions that start with `pytest`, since these functions act as
      hooks into py.test.
    - TODO: currently, all `pytest` functions are loaded into the same
      namespace, which means that if two different files imlement the same
      py.test hook, only the latter of these will be respected.
    - TODO: py.test also allows test functions to be implemented in classes;
      these tests cannot currently be loaded by this function.
    """
    tests = {}
    for module in modules:
        m = __import__('.'.join(module), globals(), locals(), ['*'])
        for k in dir(m):
            test = getattr(m, k)
            if callable(test) and re.search(pattern, k):
                args = inspect.getargspec(test).args
                if any(re.search(arg_pattern, arg) for arg in args):
                    tests['.'.join(['test'] + module + [k])] = test
            if k.startswith('pytest'):  # automatically load py.test hooks
                # TODO: different files with different implementations of the
                #   same pytest hook will break here!
                tests[k] = getattr(m, k)

    return tests


class ThreadedAssertion(object):
    """Performs assertions in parallel.

    Starts a number of threads, waits for each thread to execute some
    initialization code, and then executes assertions in each thread.
    """

    class AssertionWorker(threading.Thread):
        def __init__(self, parent, barriers, n):
            super(ThreadedAssertion.AssertionWorker, self).__init__()
            self.parent = parent
            self.barriers = barriers
            self.n = n
            self.assertion_result = None
            self.exc_info = (None, None, None)

        def run(self):
            self.parent.init_thread(self)

            self.barriers[self.n].set()
            for barrier in self.barriers:
                barrier.wait()

            try:
                self.parent.assert_thread(self)
                self.assertion_result = True
            except Exception:
                self.assertion_result = False
                self.exc_info = sys.exc_info()
            finally:
                self.parent.finish_thread(self)

    def __init__(self, n_threads):
        barriers = [threading.Event() for _ in range(n_threads)]
        threads = [self.AssertionWorker(self, barriers, i)
                   for i in range(n_threads)]
        for t in threads:
            t.start()
        for t in threads:
            t.join()
            if not t.assertion_result:
                reraise(*t.exc_info)

    def init_thread(self, worker):
        pass

    def assert_thread(self, worker):
        raise NotImplementedError()

    def finish_thread(self, worker):
        pass



import os
from . import exceptions
from . import constants


if os.name == 'nt':  # pragma: no cover
    import msvcrt

    def lock(file_, flags):
        if flags & constants.LOCK_SH:
            if flags & constants.LOCK_NB:
                mode = msvcrt.LK_NBRLCK
            else:
                mode = msvcrt.LK_RLOCK
        else:
            if flags & constants.LOCK_NB:
                mode = msvcrt.LK_NBLCK
            else:
                mode = msvcrt.LK_LOCK

        # windows locks byte ranges, so make sure to lock from file start
        try:
            savepos = file_.tell()
            file_.seek(0, os.SEEK_END)
            if file_.tell() > 0x7fffffff:
                raise exceptions.FileToLarge("Files larger than 2GB "
                                             "not supported on Windows")
            # [ ] test exclusive lock fails on seek here
            # [ ] test if shared lock passes this point
            file_.seek(0)
            try:
                # [x] check if 0 param locks entire file (not documented in
                #     Python)
                # [x] just fails with "IOError: [Errno 13] Permission denied",
                # [x] -1 crashes on py3.5+
                # [x] larger than 0x7fffffff gives OverflowError
                # [ ] writing past 2GB goes to unlocked space, should not
                #     be a problem when only using portalocker, since it
                #     always locks from the start
                msvcrt.locking(file_.fileno(), mode, 0x7fffffff)
            except IOError as exc_value:
                # [ ] be more specific here
                raise exceptions.LockException(
                    exceptions.LockException.LOCK_FAILED, exc_value.strerror)
            finally:
                if savepos:
                    file_.seek(savepos)
        except IOError as exc_value:
            raise exceptions.LockException(
                exceptions.LockException.LOCK_FAILED, exc_value.strerror)

    def unlock(file_):
        try:
            savepos = file_.tell()
            if savepos:
                file_.seek(0)
            try:
                msvcrt.locking(file_.fileno(), constants.LOCK_UN, 0x7fffffff)
            except IOError as exc_value:
                raise exceptions.LockException(
                    exceptions.LockException.LOCK_FAILED, exc_value.strerror)
            finally:
                if savepos:
                    file_.seek(savepos)
        except IOError as exc_value:
            raise exceptions.LockException(
                exceptions.LockException.LOCK_FAILED, exc_value.strerror)

elif os.name == 'posix':  # pragma: no cover
    import fcntl

    def lock(file_, flags):
        try:
            fcntl.flock(file_.fileno(), flags)
        except IOError as exc_value:
            # The exception code varies on different systems so we'll catch
            # every IO error
            raise exceptions.LockException(exc_value)

    def unlock(file_):
        fcntl.flock(file_.fileno(), constants.LOCK_UN)

else:  # pragma: no cover
    raise RuntimeError('PortaLocker only defined for nt and posix platforms')




import warnings

import numpy as np

import nengo.utils.numpy as npext
from nengo.base import NengoObject, ObjView
from nengo.exceptions import ValidationError
from nengo.params import Default, IntParam, Parameter
from nengo.processes import Process
from nengo.utils.compat import is_array_like
from nengo.utils.stdlib import checked_call


class OutputParam(Parameter):
    def __init__(self, name, default, optional=True, readonly=False):
        assert optional  # None has meaning (passthrough node)
        super(OutputParam, self).__init__(name, default, optional, readonly)

    def check_ndarray(self, node, output):
        if len(output.shape) > 1:
            raise ValidationError("Node output must be a vector (got shape "
                                  "%s)" % (output.shape,),
                                  attr=self.name, obj=node)
        if node.size_in != 0:
            raise ValidationError("output must be callable if size_in != 0",
                                  attr=self.name, obj=node)
        if node.size_out is not None and node.size_out != output.size:
            raise ValidationError("Size of Node output (%d) does not match "
                                  "size_out (%d)"
                                  % (output.size, node.size_out),
                                  attr=self.name, obj=node)

    def coerce(self, node, output):
        output = super(OutputParam, self).coerce(node, output)

        size_in_set = node.size_in is not None
        node.size_in = node.size_in if size_in_set else 0

        # --- Validate and set the new size_out
        if output is None:
            if node.size_out is not None:
                warnings.warn("'Node.size_out' is being overwritten with "
                              "'Node.size_in' since 'Node.output=None'")
            node.size_out = node.size_in
        elif isinstance(output, Process):
            if not size_in_set:
                node.size_in = output.default_size_in
            if node.size_out is None:
                node.size_out = output.default_size_out
        elif callable(output):
            # We trust user's size_out if set, because calling output
            # may have unintended consequences (e.g., network communication)
            if node.size_out is None:
                result = self.coerce_callable(node, output)
                node.size_out = 0 if result is None else result.size
        elif is_array_like(output):
            # Make into correctly shaped numpy array before validation
            output = npext.array(
                output, min_dims=1, copy=False, dtype=np.float64)
            self.check_ndarray(node, output)
            if not np.all(np.isfinite(output)):
                raise ValidationError("Output value must be finite.",
                                      attr=self.name, obj=node)
            node.size_out = output.size
        else:
            raise ValidationError("Invalid node output type %r" %
                                  type(output).__name__,
                                  attr=self.name, obj=node)

        return output

    def coerce_callable(self, node, output):
        t, x = 0.0, np.zeros(node.size_in)
        args = (t, x) if node.size_in > 0 else (t,)
        result, invoked = checked_call(output, *args)
        if not invoked:
            msg = ("output function '%s' is expected to accept exactly "
                   "%d argument" % (output, len(args)))
            msg += (' (time, as a float)' if len(args) == 1 else
                    's (time, as a float and data, as a NumPy array)')
            raise ValidationError(msg, attr=self.name, obj=node)

        if result is not None:
            result = np.asarray(result)
            if len(result.shape) > 1:
                raise ValidationError("Node output must be a vector (got shape"
                                      " %s)" % (result.shape,),
                                      attr=self.name, obj=node)
        return result


class Node(NengoObject):
    """Provide non-neural inputs to Nengo objects and process outputs.

    Nodes can accept input, and perform arbitrary computations
    for the purpose of controlling a Nengo simulation.
    Nodes are typically not part of a brain model per se,
    but serve to summarize the assumptions being made
    about sensory data or other environment variables
    that cannot be generated by a brain model alone.

    Nodes can also be used to test models by providing specific input signals
    to parts of the model, and can simplify the input/output interface of a
    `~nengo.Network` when used as a relay to/from its internal
    ensembles (see `~nengo.networks.EnsembleArray` for an example).

    Parameters
    ----------
    output : callable, array_like, or None
        Function that transforms the Node inputs into outputs,
        a constant output value, or None to transmit signals unchanged.
    size_in : int, optional (Default: 0)
        The number of dimensions of the input data parameter.
    size_out : int, optional (Default: None)
        The size of the output signal. If None, it will be determined
        based on the values of ``output`` and ``size_in``.
    label : str, optional (Default: None)
        A name for the node. Used for debugging and visualization.
    seed : int, optional (Default: None)
        The seed used for random number generation.
        Note: no aspects of the node are random, so currently setting
        this seed has no effect.

    Attributes
    ----------
    label : str
        The name of the node.
    output : callable, array_like, or None
        The given output.
    size_in : int
        The number of dimensions for incoming connection.
    size_out : int
        The number of output dimensions.
    """

    probeable = ('output',)

    output = OutputParam('output', default=None)
    size_in = IntParam('size_in', default=None, low=0, optional=True)
    size_out = IntParam('size_out', default=None, low=0, optional=True)

    def __init__(self, output=Default, size_in=Default, size_out=Default,
                 label=Default, seed=Default):
        if not (seed is Default or seed is None):
            raise NotImplementedError(
                "Changing the seed of a node has no effect")
        super(Node, self).__init__(label=label, seed=seed)

        self.size_in = size_in
        self.size_out = size_out
        self.output = output  # Must be set after size_out; may modify size_out

    def __getitem__(self, key):
        return ObjView(self, key)

    def __len__(self):
        return self.size_out



from __future__ import absolute_import

import logging
import sys

from .compat import TextIO

console_formatter = logging.Formatter('[%(levelname)s] %(message)s')
file_formatter = logging.Formatter(
    ('%(asctime)s [%(levelname)s] %(name)s.%(funcName)s'
     '@ L%(lineno)d\n  %(message)s'))
console_handler = logging.StreamHandler(sys.stdout)
console_handler.setFormatter(console_formatter)


def log(level='warning', path=None):
    """Log messages.

    If path is None, logging messages will be printed to the console (stdout).
    If it not None, logging messages will be appended to the file at that path.

    Typically someone using Nengo as a library will set up their own
    logging things, and Nengo will just populate their log.
    However, if the user is using Nengo directly, they can use this
    function to get log output.

    Parameters
    ----------
    level : string (optional)
        Collect messages targeting this level and above. The levels from
        highest to lowest are: 'CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG'.
    path : string (optional)
        Path of a file to append log messages to. If ``None`` (default),
        messages are logged to the console.

    Returns
    -------
    logging handler
        The logging handler setup by this function.
    """
    if level.upper() not in ('CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG'):
        raise ValueError("Invalid logging level")

    level = getattr(logging, level.upper())
    logging.root.setLevel(level)

    if path is None:
        handler = console_handler
    else:
        for handler in logging.root.handlers:
            if (isinstance(handler, logging.FileHandler)
                    and handler.baseFilename == path
                    and handler.formatter == file_formatter):
                break
        else:
            handler = logging.FileHandler(path, encoding='utf-8')
            handler.setFormatter(file_formatter)

    if handler not in logging.root.handlers:
        logging.root.addHandler(handler)
    handler.setLevel(level)
    logging.captureWarnings(True)
    return handler


class CaptureLogHandler(logging.StreamHandler):
    """A logging handler that stores log records and the log text."""

    def __init__(self):
        """Creates a new log handler."""
        logging.StreamHandler.__init__(self)
        self.records = []
        self.stream = TextIO()

    def close(self):
        """Close this log handler and its underlying stream."""
        self.stream.close()
        logging.StreamHandler.close(self)

    def emit(self, record):
        """Keep the log records in a list in addition to the log text."""
        self.records.append(record)
        logging.StreamHandler.emit(self, record)



"""
Networks (:mod:`nengolib.networks`)
===================================

.. currentmodule:: nengolib.networks

.. autosummary::
   :toctree:

   LinearNetwork
   RollingWindow
"""

from .echo_state import *
from .linear_network import *
from .reservoir import *
from .rolling_window import *



import numpy as np
import pytest

import nengo
from nengo.builder import Model
from nengo.builder.ensemble import BuiltEnsemble
from nengo.builder.signal import Signal, SignalDict
from nengo.exceptions import ObsoleteError, SignalError
from nengo.utils.compat import itervalues, range


def test_seeding(RefSimulator, logger):
    """Test that setting the model seed fixes everything"""

    #  TODO: this really just checks random parameters in ensembles.
    #   Are there other objects with random parameters that should be
    #   tested? (Perhaps initial weights of learned connections)

    m = nengo.Network(label="test_seeding")
    with m:
        input = nengo.Node(output=1, label="input")
        A = nengo.Ensemble(40, 1, label="A")
        B = nengo.Ensemble(20, 1, label="B")
        nengo.Connection(input, A)
        C = nengo.Connection(A, B, function=lambda x: x ** 2)

    m.seed = 872
    with RefSimulator(m) as sim:
        m1 = sim.data
    with RefSimulator(m) as sim:
        m2 = sim.data
    m.seed = 873
    with RefSimulator(m) as sim:
        m3 = sim.data

    def compare_objs(obj1, obj2, attrs, equal=True):
        for attr in attrs:
            check = (np.allclose(getattr(obj1, attr), getattr(obj2, attr)) ==
                     equal)
            if not check:
                logger.info("%s: %s", attr, getattr(obj1, attr))
                logger.info("%s: %s", attr, getattr(obj2, attr))
            assert check

    ens_attrs = BuiltEnsemble._fields
    As = [mi[A] for mi in [m1, m2, m3]]
    Bs = [mi[B] for mi in [m1, m2, m3]]
    compare_objs(As[0], As[1], ens_attrs)
    compare_objs(Bs[0], Bs[1], ens_attrs)
    compare_objs(As[0], As[2], ens_attrs, equal=False)
    compare_objs(Bs[0], Bs[2], ens_attrs, equal=False)

    conn_attrs = ('eval_points', 'weights')
    Cs = [mi[C] for mi in [m1, m2, m3]]
    compare_objs(Cs[0], Cs[1], conn_attrs)
    compare_objs(Cs[0], Cs[2], conn_attrs, equal=False)


def test_hierarchical_seeding():
    """Changes to subnetworks shouldn't affect seeds in top-level network"""

    def create(make_extra, seed):
        objs = []
        with nengo.Network(seed=seed, label='n1') as model:
            objs.append(nengo.Ensemble(10, 1, label='e1'))
            with nengo.Network(label='n2'):
                objs.append(nengo.Ensemble(10, 1, label='e2'))
                if make_extra:
                    # This shouldn't affect any seeds
                    objs.append(nengo.Ensemble(10, 1, label='e3'))
            objs.append(nengo.Ensemble(10, 1, label='e4'))
        return model, objs

    same1, same1objs = create(False, 9)
    same2, same2objs = create(True, 9)
    diff, diffobjs = create(True, 10)

    m1 = Model()
    m1.build(same1)
    same1seeds = m1.seeds

    m2 = Model()
    m2.build(same2)
    same2seeds = m2.seeds

    m3 = Model()
    m3.build(diff)
    diffseeds = m3.seeds

    for diffobj, same2obj in zip(diffobjs, same2objs):
        # These seeds should all be different
        assert diffseeds[diffobj] != same2seeds[same2obj]

    # Skip the extra ensemble
    same2objs = same2objs[:2] + same2objs[3:]

    for same1obj, same2obj in zip(same1objs, same2objs):
        # These seeds should all be the same
        assert same1seeds[same1obj] == same2seeds[same2obj]


def test_signal():
    """Make sure assert_named_signals works."""
    Signal(np.array(0.))
    Signal.assert_named_signals = True
    with pytest.raises(AssertionError):
        Signal(np.array(0.))

    # So that other tests that build signals don't fail...
    Signal.assert_named_signals = False


def test_signal_values():
    """Make sure Signal.initial_value works."""
    two_d = Signal([[1.], [1.]])
    assert np.allclose(two_d.initial_value, np.array([[1], [1]]))
    two_d_view = two_d[0, :]
    assert np.allclose(two_d_view.initial_value, np.array([1]))

    # cannot change signal value after creation
    with pytest.raises(SignalError):
        two_d.initial_value = np.array([[0.5], [-0.5]])
    with pytest.raises((ValueError, RuntimeError)):
        two_d.initial_value[...] = np.array([[0.5], [-0.5]])


def test_signaldict():
    """Tests SignalDict's dict overrides."""
    signaldict = SignalDict()

    scalar = Signal(1.)

    # Both __getitem__ and __setitem__ raise KeyError
    with pytest.raises(KeyError):
        signaldict[scalar]
    with pytest.raises(KeyError):
        signaldict[scalar] = np.array(1.)

    signaldict.init(scalar)
    assert np.allclose(signaldict[scalar], np.array(1.))
    # __getitem__ handles scalars
    assert signaldict[scalar].shape == ()

    one_d = Signal([1.])
    signaldict.init(one_d)
    assert np.allclose(signaldict[one_d], np.array([1.]))
    assert signaldict[one_d].shape == (1,)

    two_d = Signal([[1.], [1.]])
    signaldict.init(two_d)
    assert np.allclose(signaldict[two_d], np.array([[1.], [1.]]))
    assert signaldict[two_d].shape == (2, 1)

    # __getitem__ handles views implicitly (note no .init)
    two_d_view = two_d[0, :]
    assert np.allclose(signaldict[two_d_view], np.array([1.]))
    assert signaldict[two_d_view].shape == (1,)

    # __setitem__ ensures memory location stays the same
    memloc = signaldict[scalar].__array_interface__['data'][0]
    signaldict[scalar] = np.array(0.)
    assert np.allclose(signaldict[scalar], np.array(0.))
    assert signaldict[scalar].__array_interface__['data'][0] == memloc

    memloc = signaldict[one_d].__array_interface__['data'][0]
    signaldict[one_d] = np.array([0.])
    assert np.allclose(signaldict[one_d], np.array([0.]))
    assert signaldict[one_d].__array_interface__['data'][0] == memloc

    memloc = signaldict[two_d].__array_interface__['data'][0]
    signaldict[two_d] = np.array([[0.], [0.]])
    assert np.allclose(signaldict[two_d], np.array([[0.], [0.]]))
    assert signaldict[two_d].__array_interface__['data'][0] == memloc

    # __str__ pretty-prints signals and current values
    # Order not guaranteed for dicts, so we have to loop
    for k in signaldict:
        assert "%s %s" % (repr(k), repr(signaldict[k])) in str(signaldict)


def test_signaldict_reset():
    """Tests SignalDict's reset function."""
    signaldict = SignalDict()
    two_d = Signal([[1.], [1.]])
    signaldict.init(two_d)

    two_d_view = two_d[0, :]
    signaldict.init(two_d_view)

    signaldict[two_d_view] = -0.5
    assert np.allclose(signaldict[two_d], np.array([[-0.5], [1]]))

    signaldict[two_d] = np.array([[-1], [-1]])
    assert np.allclose(signaldict[two_d], np.array([[-1], [-1]]))
    assert np.allclose(signaldict[two_d_view], np.array([-1]))

    signaldict.reset(two_d_view)
    assert np.allclose(signaldict[two_d_view], np.array([1]))
    assert np.allclose(signaldict[two_d], np.array([[1], [-1]]))

    signaldict.reset(two_d)
    assert np.allclose(signaldict[two_d], np.array([[1], [1]]))


def test_signal_reshape():
    """Tests Signal.reshape"""
    three_d = Signal(np.ones((2, 2, 2)))
    assert three_d.reshape((8,)).shape == (8,)
    assert three_d.reshape((4, 2)).shape == (4, 2)
    assert three_d.reshape((2, 4)).shape == (2, 4)
    assert three_d.reshape(-1).shape == (8,)
    assert three_d.reshape((4, -1)).shape == (4, 2)
    assert three_d.reshape((-1, 4)).shape == (2, 4)
    assert three_d.reshape((2, -1, 2)).shape == (2, 2, 2)
    assert three_d.reshape((1, 2, 1, 2, 2, 1)).shape == (1, 2, 1, 2, 2, 1)


def test_signal_slicing(rng):
    slices = [0, 1, slice(None, -1), slice(1, None), slice(1, -1),
              slice(None, None, 3), slice(1, -1, 2)]

    x = np.arange(12, dtype=float)
    y = np.arange(24, dtype=float).reshape(4, 6)
    a = Signal(x.copy())
    b = Signal(y.copy())

    for i in range(100):
        si0, si1 = rng.randint(0, len(slices), size=2)
        s0, s1 = slices[si0], slices[si1]
        assert np.array_equiv(a[s0].initial_value, x[s0])
        assert np.array_equiv(b[s0, s1].initial_value, y[s0, s1])

    with pytest.raises(ValueError):
        a[[0, 2]]
    with pytest.raises(ValueError):
        b[[0, 1], [3, 4]]


def test_commonsig_readonly():
    """Test that the common signals cannot be modified."""
    net = nengo.Network(label="test_commonsig")
    model = Model()
    model.build(net)
    signals = SignalDict()

    for sig in itervalues(model.sig['common']):
        signals.init(sig)
        with pytest.raises((ValueError, RuntimeError)):
            signals[sig] = np.array([-1])
        with pytest.raises((ValueError, RuntimeError)):
            signals[sig][...] = np.array([-1])


def test_obsolete_params(RefSimulator):
    with nengo.Network() as net:
        e = nengo.Ensemble(10, 1)
        c = nengo.Connection(e, e)
    with RefSimulator(net) as sim:
        pass
    with pytest.raises(ObsoleteError):
        sim.data[c].decoders



from .actionselection import BasalGanglia, Thalamus
from .assoc_mem import AssociativeMemory
from .circularconvolution import CircularConvolution
from .ensemblearray import EnsembleArray
from .integrator import Integrator
from .oscillator import Oscillator
from .product import Product
from .workingmemory import InputGatedMemory



# -*- coding: utf-8 -*-

import warnings

import numpy as np

from scipy.linalg import inv
from scipy.special import binom

import nengo
from nengo.params import Default
from nengo.exceptions import ValidationError
from nengo.utils.network import with_self
from nengo.utils.stdlib import checked_call

from nengolib.networks.linear_network import LinearNetwork
from nengolib.signal.dists import EvalPoints, Encoders
from nengolib.signal.realizers import Balanced
from nengolib.synapses.analog import PadeDelay, LegendreDelay

__all__ = ['t_default', 'RollingWindow']

t_default = np.linspace(0, 1, 1000)  # default window time points (normalized)


def _pade_readout(q, r):
    """C matrix to decode a delay of r*theta from PadeDelay.

    ``r`` is a ratio between 0 (``t=0``) and 1 (``t=-theta``).
    """

    c = np.zeros(q)
    for i in range(q):
        j = np.arange(i+1, dtype=np.float64)
        c[q-1-i] += 1 / binom(q, i) * np.sum(
            binom(q, j) * binom(2*q - 1 - j, i - j) * (-r)**(i - j))
    return c


def _legendre_readout(q, r):
    """C matrix to decode a delay of r*theta from LegendreDelay.

    ``r`` is a ratio between 0 (``t=0``) and 1 (``t=-theta``).
    """

    from scipy.special import legendre
    return np.asarray([legendre(i)(2*r - 1) for i in range(q)])


class RollingWindow(LinearNetwork):
    """Approximate nonlinear functions across a rolling window of input.

    Compresses the input history of finite length into a low-dimensional state
    to support the approximation of nonlinear functions across a rolling
    window of input history.  This can be used to approximate
    FIR filters and window functions in continuous time.

    Parameters
    ----------
    theta : ``float``
        Width of rolling window (time-delay) in seconds.
    n_neurons : ``integer``
        Total number of neurons to use in :class:`nengo.Ensemble`.
    process : :class:`nengo.Process`
        Process modelling a typical input to the network.
        Used to optimize the lengths of axis-aligned ``encoders`` and
        the distribution of ``eval_points``.
        If ``None``, then will use the defaults of uniform unit-length
        ``encoders`` and uniformly distributed ``eval_points``, which
        usually gives sub-optimal performance.
    dimensions : ``integer``, optional
        Order of :func:`.PadeDelay`, or dimensionality of the state vector.
        Defaults to ``6``, which has an approximation error of less than
        one percent for input frequencies less than ``1/theta``
        (see :func:`.pade_delay_error`).
    dt : ``float``, optional
        Simulation time-step (in seconds), passed to :class:`.LinearNetwork`.
        Defaults to ``0.001``, but should be manually specified if the
        simulation time-step is ever changed.
    synapse : :class:`nengo.synapses.Synapse`, optional
        Recurrent synapse. Bigger is typically better.
        Defaults to ``0.1``.
    input_synapse : :class:`nengo.synapses.Synapse`, optional
        Input synapse. Typically should match the value of
        ``synapse``, unless the input has already been filtered by the
        same synapse. Defaults to ``0.1``.
    realizer : :class:`.AbstractRealizer`, optional
        Method of realizing the linear system. Defaults to :class:`.Balanced`.
    solver : :class:`nengo.solvers.Solver`, optional
        Solver to use for connections from the state ensemble.
        Defaults to :class:`nengo.solvers.LstsqL2` (with ``reg=1e-3``).
    legendre : ``boolean``, optional
        Changes the canonical realization to that of :func:`.LegendreDelay`,
        the shifted Legendre polynomials,
        :math:`P_i(2 \\theta' \\theta^{-1} - 1)`.
        Defaults to ``False``.
    **kwargs : ``dictionary``, optional
        Additional keyword arguments passed to :class:`.LinearNetwork`.
        Those that fall under the heading of ``**ens_kwargs`` will be
        passed to the :class:`nengo.Ensemble` that represents the
        :attr:`.state`.

    See Also
    --------
    :class:`.LinearNetwork`
    :func:`.PadeDelay`
    :func:`.pade_delay_error`
    :class:`.Balanced`
    :class:`.EvalPoints`
    :class:`.Encoders`

    Notes
    -----
    This extends :class:`.LinearNetwork` to efficiently implement the
    :func:`.PadeDelay` system (via :class:`.Balanced`, :class:`.EvalPoints`,
    and :class:`.Encoders`), with support for decoding nonlinear functions
    from the state of the network. [#]_

    Function are decoded from the window by evaluating the state-space
    at arbitrary points (as usual in Nengo), and linearly projecting them onto
    basis functions to express each point as a sampled window representation.
    See :func:`.add_output` for details.

    References
    ----------
    .. [#] A. R. Voelker and C. Eliasmith, "Improving spiking dynamical
       networks: Accurate delays, higher-order synapses, and time cells",
       Neural Computation (preprint), accepted 09 2017.
       [`URL <https://github.com/arvoelke/delay2017>`__]

    Examples
    --------
    See :doc:`notebooks/examples/rolling_window` for a notebook example.

    >>> from nengolib.networks import RollingWindow, t_default

    Approximate the maximum of a window of width 50 ms, as well as a sampling
    of the window itself. The :class:`.Hankel` realizer happens to be better
    than the default of :class:`.Balanced` for computing the ``max`` function.

    >>> import nengo
    >>> from nengolib import Network
    >>> from nengolib.signal import Hankel
    >>> with Network() as model:
    >>>     process = nengo.processes.WhiteSignal(100., high=25, y0=0)
    >>>     stim = nengo.Node(output=process)
    >>>     rw = RollingWindow(theta=.05, n_neurons=2500, process=process,
    >>>                        neuron_type=nengo.LIFRate(), legendre=True)
    >>>     nengo.Connection(stim, rw.input, synapse=None)
    >>>     p_stim = nengo.Probe(stim)
    >>>     p_delay = nengo.Probe(rw.output)
    >>>     p_max = nengo.Probe(rw.add_output(function=np.max))
    >>>     p_window = nengo.Probe(rw.add_output(function=lambda w: w[::20]))
    >>> with nengo.Simulator(model, seed=0) as sim:
    >>>     sim.run(.5)

    >>> import matplotlib.pyplot as plt
    >>> plt.subplot(211)
    >>> plt.plot(sim.trange(), sim.data[p_stim], label="Input")
    >>> plt.plot(sim.trange(), sim.data[p_delay], label="Delay")
    >>> plt.legend()
    >>> plt.subplot(212)
    >>> plt.plot(sim.trange(), sim.data[p_window], alpha=.2)
    >>> plt.plot(sim.trange(), sim.data[p_max], c='purple', label="max(w)")
    >>> plt.legend()
    >>> plt.xlabel("Time (s)")
    >>> plt.show()

    Visualizing the canonical basis functions. The state of the
    :func:`LegendreDelay` system takes a linear combination of these
    shifted Legendre polynomials to represent the current window of history:

    >>> plt.title("canonical_basis()")
    >>> plt.plot(t_default, rw.canonical_basis())
    >>> plt.xlabel("Normalized Time (Unitless)")
    >>> plt.show()

    Visualizing the realized basis functions. This is a linear transformation
    of the above basis functions according to the realized state-space
    (see ``realizer`` parameter). The state of the **network** takes a linear
    combination of these to represent the current window of history:

    >>> plt.title("basis()")
    >>> plt.plot(t_default, rw.basis())
    >>> plt.xlabel("Normalized Time (Unitless)")
    >>> plt.show()

    Visualizing the inverse basis functions. The functions that can be
    accurately decoded are expressed in terms of the dot-product of the window
    with these functions (see :func:`.add_output` for mathematical details).

    >>> plt.title("inverse_basis().T")
    >>> plt.plot(t_default, rw.inverse_basis().T)
    >>> plt.xlabel("Normalized Time (Unitless)")
    >>> plt.show()
    """

    # TODO: see notebook for more details.

    def __init__(self, theta, n_neurons, process, dimensions=6,
                 synapse=0.1, input_synapse=0.1, dt=0.001, realizer=Balanced(),
                 solver=nengo.solvers.LstsqL2(reg=1e-3), legendre=False,
                 **kwargs):
        self.theta = theta
        self.process = process
        self.dimensions = dimensions
        self.legendre = legendre

        if self.legendre:
            self._readout = _legendre_readout
            sys = LegendreDelay(theta, order=dimensions)

        else:
            self._readout = _pade_readout
            sys = PadeDelay(theta, order=dimensions)

        super(RollingWindow, self).__init__(
            sys=sys,
            n_neurons_per_ensemble=n_neurons,
            input_synapse=input_synapse,
            synapse=synapse,
            dt=dt,
            realizer=realizer,
            solver=solver,
            **kwargs)

    def _make_core(self, solver, **ens_kwargs):
        if self.process is not None:  # set by RollingWindow.__init__
            for illegal in ('eval_points', 'encoders', 'normalize_encoders'):
                if illegal in ens_kwargs and \
                   ens_kwargs[illegal] is not Default:
                    raise ValidationError(
                        "'%s' must not be given (%s) if 'process' is not "
                        "None." % (illegal, ens_kwargs[illegal]),
                        attr=illegal, obj=self)

            # Wrap the realized state by an eval_point and encoder process
            # This is done here automatically for convenience, but finer-grain
            # control can be achieved by passing in your own eval_points and
            # encoders and keeping process=None. These can also be set
            # directly on self.state after initialization, and any previously
            # created connections will still inherit the new eval_points.
            X = self.realization.X  # set by LinearNetwork
            ens_kwargs['eval_points'] = EvalPoints(X, self.process, dt=self.dt)
            ens_kwargs['encoders'] = Encoders(X, self.process, dt=self.dt)
            if nengo.version.version_info >= (2, 4, 0):
                ens_kwargs['normalize_encoders'] = False

            else:  # pragma: no cover
                warnings.warn(
                    "'normalize_encoders' is not supported by nengo<=%s, and "
                    "so the 'radii' for the representation cannot be "
                    "automatically optimized; try tweaking the 'radii' "
                    "manually, or upgrading to nengo>=2.4.0." %
                    nengo.__version__, UserWarning)

        self._state = nengo.Ensemble(
            n_neurons=self.n_neurons_per_ensemble,
            dimensions=self.size_state,
            label="state",
            **ens_kwargs)

        # For plausibility, only linear transformations should be made from
        # the output node. Nonlinear transformations should be decoded from
        # the state via self.add_output(...).
        output = nengo.Node(size_in=self.size_state)
        nengo.Connection(self.state, output, synapse=None, solver=solver)
        return self.state, output

    def canonical_basis(self, t=t_default):
        """Temporal basis functions for PadeDelay in canonical form."""
        t = np.atleast_1d(t)
        B = np.asarray([self._readout(self.dimensions, r) for r in t])
        return B

    def basis(self, t=t_default):
        """Temporal basis functions for realized PadeDelay."""
        # Undo change of basis from realizer, and then transform into window
        B = self.canonical_basis(t)
        return B.dot(self.realizer_result.T)

    def inverse_basis(self, t=t_default):
        """Moore-Penrose pseudoinverse of the basis functions."""
        B = self.basis(t)
        return inv(B.T.dot(B)).dot(B.T)

    @with_self
    def add_output(self, t=None, function=lambda w: w[-1], label='output',
                   synapse=None, **conn_kwargs):
        """Decodes a function of the window at time points ``-t*theta``.

        Parameters
        ----------
        t : ``array_like``, optional
            A scalar or array-like with elements ranging between ``0``
            (beginning of window) and ``1`` (end of window; i.e., ``theta``).
            Specifies the time-points at which to sample the window's basis
            functions. Defaults to the value of :data:`.t_default`:
            ``1000`` points spaced evenly between ``0`` and ``1``.
        function : ``callable``, optional
            A function that consumes some window ``w`` of requested time-points
            (``len(w) == len(t)``), and returns the desired decoding.
            Defaults to returning the end of the window: ``w[-1]``.
        label : ``string``, optional
            Label for the created :class:`nengo.Node`.
            Defaults to ``'output'``.
        synapse : :class:`nengo.synapses.Synapse`, optional
            Synapse passed to created :class:`nengo.Connection`.
            Defaults to ``None``.
        **conn_kwargs : ``dictionary``, optional
            Additional keyword arguments passed to :class:`nengo.Connection`.

        Returns
        -------
        :class:`nengo.Node`
            Node object that holds the decoded function from the state
            ensemble. The size of the node is equal to the output
            dimensionality of the provided function.

        Notes
        -----
        The approach is to project the state-vector onto the basis functions
        given by the rows of :func:`.basis`, and then supply the resulting
        window representation to the given ``function``. Then we solve for
        the decoders with respect to the state-vector as usual in Nengo.

        The :func:`.canonical_basis` functions are polynomials of increasing
        order (from ``0`` up to ``q-1``, where ``q=dimensions``).
        If ``legendre == False``, then:

        .. math::

            P_i(t) = \\begin{pmatrix}q \\\\ i\\end{pmatrix}^{-1}
                \\sum_{j=0}^i \\begin{pmatrix}q \\\\ j\\end{pmatrix}
                \\begin{pmatrix}2q - 1 - j \\\\ i - j\\end{pmatrix}
                \\left( -t \\right)^{i - j}
                \\text{,} \\quad 0 \\le t \\le 1
                \\text{,} \\quad i = 0 \\ldots q - 1 \\text{.}

        If ``legendre == True``, then these are the shifted Legendre
        polynomials: [#]_

        .. math::

            P_i(t) = (-1)^i
                \\sum_{j=0}^i \\begin{pmatrix}i \\\\ j\\end{pmatrix}
                \\begin{pmatrix}i + j \\\\ j\\end{pmatrix}
                \\left( -t \\right)^{j}
                \\text{,} \\quad 0 \\le t \\le 1
                \\text{,} \\quad i = 0 \\ldots q - 1 \\text{.}

        Since the encoders are axis-aligned (when a ``process`` is given),
        the functions that can be accurately decoded by this approach are of
        the form:

        .. math::

           f({\\bf w}) = \\sum_{i=0}^{q-1} f_i ({\\bf v}_i \\cdot {\\bf w})

        where :math:`{\\bf w}` is some history, :math:`{\\bf v}_i` are the
        columns of :func:`.inverse_basis`, and each :math:`f_i` is some
        unknown low-degree nonlinearity.

        References
        ----------
        .. [#] Rodrigues, Olinde. De l'attraction des sphéroïdes,
           Correspondence sur l'É-cole Impériale Polytechnique.
           Diss. Thesis for the Faculty of Science of the University of Paris)
           3 (3): 361–385, 1816.
        """

        if t is None:
            t = t_default

        B = self.basis(t)

        def wrapped_function(x):
            w = B.dot(x)  # state -> window
            return function(w)

        value, invoked = checked_call(
            function, np.zeros(B.shape[0]))
        if not invoked:
            raise ValidationError(
                "'function' (%s) must accept a single np.array argument of "
                "size=%d." % (function, B.shape[0]),
                attr='function', obj=self)

        output = nengo.Node(size_in=np.asarray(value).size, label=label)
        nengo.Connection(self.state, output, function=wrapped_function,
                         synapse=synapse, **conn_kwargs)
        return output



import numpy as np
import pytest

import nengo

from nengolib.compat import get_activities
from nengolib.signal.learning import pes_learning_rate
from nengolib import Network


def test_pes_learning_rate(Simulator, plt, seed):
    n = 50
    dt = 0.0005
    T = 1.0
    initial = 0.7
    desired = -0.9
    epsilon = 1e-3  # get to factor epsilon with T seconds

    # Get activity vector and initial decoders
    with Network(seed=seed) as model:
        x = nengo.Ensemble(
            n, 1, seed=seed, neuron_type=nengo.neurons.LIFRate())
        y = nengo.Node(size_in=1)
        conn = nengo.Connection(x, y, synapse=None)

    with Simulator(model, dt=dt) as sim:
        a = get_activities(sim.model, x, [initial])
        d = sim.data[conn].weights
        assert np.any(a > 0)

    # Use util function to calculate learning_rate
    init_error = float(desired - np.dot(d, a))
    learning_rate, gamma = pes_learning_rate(
        epsilon / abs(init_error), a, T, dt)

    # Build model with no filtering on any connections
    with model:
        stim = nengo.Node(output=initial)
        ystar = nengo.Node(output=desired)

        conn.learning_rule_type = nengo.PES(
            pre_tau=1e-15, learning_rate=learning_rate)

        nengo.Connection(stim, x, synapse=None)
        nengo.Connection(ystar, conn.learning_rule, synapse=0, transform=-1)
        nengo.Connection(y, conn.learning_rule, synapse=0)

        p = nengo.Probe(y, synapse=None)
        decoders = nengo.Probe(conn, 'weights', synapse=None)

    with Simulator(model, dt=dt) as sim:
        sim.run(T)

    # Check that the final error is exactly epsilon
    assert np.allclose(abs(desired - sim.data[p][-1]), epsilon)

    # Check that all of the errors are given exactly by gamma**k
    k = np.arange(len(sim.trange()))
    error = init_error * gamma ** k
    assert np.allclose(sim.data[p].flatten(), desired - error)

    # Check that all of the decoders are equal to their analytical solution
    dk = d.T + init_error * a.T[:, None] * (1 - gamma ** k) / np.dot(a, a.T)
    assert np.allclose(dk, np.squeeze(sim.data[decoders].T))

    plt.figure()
    plt.plot(sim.trange(), sim.data[p], lw=5, alpha=0.5)
    plt.plot(sim.trange(), desired - error, linestyle='--', lw=5, alpha=0.5)


def test_pes_learning_rate_fail():
    with pytest.raises(ValueError):
        pes_learning_rate(0.1, [[1, 2], [3, 4]], 1)



"""Utilities for progress tracking and display to the user."""

from __future__ import absolute_import, division

from datetime import timedelta
import importlib
import os
import sys
import threading
import time
import uuid
import warnings

import numpy as np

from .compat import escape
from .stdlib import get_terminal_size
from .ipython import check_ipy_version, get_ipython
from ..exceptions import ValidationError
from ..rc import rc


if get_ipython() is not None:
    from IPython.display import display, Javascript


class MemoryLeakWarning(UserWarning):
    pass


warnings.filterwarnings('once', category=MemoryLeakWarning)


def timestamp2timedelta(timestamp):
    if timestamp == -1:
        return "Unknown"
    return timedelta(seconds=np.ceil(timestamp))


def _load_class(name):
    mod_name, cls_name = name.rsplit('.', 1)
    mod = importlib.import_module(mod_name)
    return getattr(mod, cls_name)


class Progress(object):
    """Stores and tracks information about the progress of some process.

    This class is to be used as part of a ``with`` statement. Use ``step()`` to
    update the progress.

    Parameters
    ----------
    max_steps : int
        The total number of calculation steps of the process.
    name_during : str, optional
        Short description of the task to be used while it is running.
    name_after : str, optional
        Short description of the task to be used after it has
        finished. Defaults to ``name_during``.

    Attributes
    ----------
    max_steps : int, optional
        The total number of calculation steps of the process, if known.
    name_after : str
        Name of the task to be used after it has finished.
    name_during : str
        Name of the task to be used while it is running.
    steps : int
        Number of completed steps.
    success : bool or None
        Whether the process finished successfully. ``None`` if the process
        did not finish yet.
    time_end : float
        Time stamp of the time the process was finished or aborted.
    time_start : float
        Time stamp of the time the process was started.

    Examples
    --------

    >>> max_steps = 10
    >>> with Progress(max_steps=max_steps) as progress:
    ...     for i in range(max_steps):
    ...         # do something
    ...         progress.step()

    """

    def __init__(self, name_during='', name_after=None, max_steps=None):
        if max_steps is not None and max_steps <= 0:
            raise ValidationError("must be at least 1 (got %d)"
                                  % (max_steps,), attr="max_steps")
        self.n_steps = 0
        self.max_steps = max_steps
        self.name_during = name_during
        if name_after is None:
            name_after = name_during
        self.name_after = name_after
        self.time_start = self.time_end = time.time()
        self.finished = False
        self.success = None

    @property
    def progress(self):
        """The current progress as a number from 0 to 1 (inclusive).

        Returns
        -------
        float
        """
        if self.max_steps is None:
            return 0.
        return min(1.0, self.n_steps / self.max_steps)

    def elapsed_seconds(self):
        """The number of seconds passed since entering the ``with`` statement.

        Returns
        -------
        float
        """
        if self.finished:
            return self.time_end - self.time_start
        else:
            return time.time() - self.time_start

    def eta(self):
        """The estimated number of seconds until the process is finished.

        Stands for estimated time of arrival (ETA).
        If no estimate is available -1 will be returned.

        Returns
        -------
        float
        """
        if self.progress > 0.:
            return (
                (1. - self.progress) * self.elapsed_seconds() / self.progress)
        else:
            return -1

    def __enter__(self):
        self.finished = False
        self.success = None
        self.n_steps = 0
        self.time_start = time.time()
        return self

    def __exit__(self, exc_type, dummy_exc_value, dummy_traceback):
        self.success = exc_type is None
        if self.success and self.max_steps is not None:
            self.n_steps = self.max_steps
        self.time_end = time.time()
        self.finished = True

    def step(self, n=1):
        """Advances the progress.

        Parameters
        ----------
        n : int
            Number of steps to advance the progress by.
        """
        self.n_steps += n


class ProgressBar(object):
    """Visualizes the progress of a process.

    This is an abstract base class that progress bar classes some inherit from.
    Progress bars should visually displaying the progress in some way.
    """

    supports_fast_ipynb_updates = False

    def update(self, progress):
        """Updates the displayed progress.

        Parameters
        ----------
        progress : :class:`Progress`
            The progress information to display.
        """
        raise NotImplementedError()

    def close(self):
        """Closes the progress bar.

        Indicates that not further updates will be made.
        """
        pass


class NoProgressBar(ProgressBar):
    """A progress bar that does not display anything.

    Helpful in headless situations or when using Nengo as a library.
    """

    def update(self, progress):
        pass


class TerminalProgressBar(ProgressBar):
    """A progress bar that is displayed as ASCII output on ``stdout``."""

    def update(self, progress):
        if progress.finished:
            line = self._get_finished_line(progress)
        elif progress.max_steps is None:
            line = self._get_unknown_progress_line(progress)
        else:
            line = self._get_in_progress_line(progress)
        sys.stdout.write(line)
        sys.stdout.flush()

    def _get_in_progress_line(self, progress):
        line = "[{{}}] ETA: {eta}".format(
            eta=timestamp2timedelta(progress.eta()))
        percent_str = " {}... {}% ".format(
            progress.name_during, int(100 * progress.progress))
        width, _ = get_terminal_size()
        progress_width = max(0, width - len(line))
        progress_str = (
            int(progress_width * progress.progress) * "#").ljust(
                progress_width)

        percent_pos = (len(progress_str) - len(percent_str)) // 2
        if percent_pos > 0:
            progress_str = (
                progress_str[:percent_pos] + percent_str +
                progress_str[percent_pos + len(percent_str):])

        return '\r' + line.format(progress_str)

    def _get_unknown_progress_line(self, progress):
        """Generates a progress line with continuously moving marker.

        This is to indicate processing while not knowing how far along we
        progressed with the processing.
        """
        duration = progress.elapsed_seconds()
        line = "[{{}}] duration: {duration}".format(
            duration=timestamp2timedelta(duration))
        text = " {}... ".format(progress.name_during)
        width, _ = get_terminal_size()
        marker = '>>>>'
        progress_width = max(0, width - len(line) + 2)
        index_width = progress_width + len(marker)
        i = int(10. * duration) % (index_width + 1)
        progress_str = (' ' * i) + marker + (' ' * (index_width - i))
        progress_str = progress_str[len(marker):-len(marker)]
        text_pos = (len(progress_str) - len(text)) // 2
        progress_str = (
            progress_str[:text_pos] + text +
            progress_str[text_pos + len(text):])
        return '\r' + line.format(progress_str)

    def _get_finished_line(self, progress):
        width, _ = get_terminal_size()
        line = "{} finished in {}.".format(
            progress.name_after,
            timestamp2timedelta(progress.elapsed_seconds())).ljust(width)
        return '\r' + line

    def close(self):
        sys.stdout.write(os.linesep)
        sys.stdout.flush()


class HtmlProgressBar(ProgressBar):
    """A progress bar using a HTML representation.

    This HTML representation can be used in Jupyter notebook environments
    and is provided by the *_repr_html_* method that will be automatically
    used by IPython interpreters.

    If the kernel frontend does not support HTML (e.g., in Jupyter qtconsole),
    a warning message will be issued as the ASCII representation.
    """
    supports_fast_ipynb_updates = True

    def __init__(self):
        super(HtmlProgressBar, self).__init__()
        self._uuid = uuid.uuid4()
        self._handle = None

    def update(self, progress):
        if self._handle is None:
            display(self._HtmlBase(self._uuid))
            self._handle = display(self._js_update(progress), display_id=True)
        else:
            self._handle.update(self._js_update(progress))

    class _HtmlBase(object):
        def __init__(self, uuid):
            self.uuid = uuid

        def __repr__(self):
            return (
                "HtmlProgressBar cannot be displayed. Please use the "
                "TerminalProgressBar. It can be enabled with "
                "`nengo.rc.set('progress', 'progress_bar', "
                "'nengo.utils.progress.TerminalProgressBar')`.")

        def _repr_html_(self):
            return '''
                <div id="{uuid}" style="
                    width: 100%;
                    border: 1px solid #cfcfcf;
                    border-radius: 4px;
                    text-align: center;
                    position: relative;">
                  <div class="pb-text" style="
                      position: absolute;
                      width: 100%;">
                    0%
                  </div>
                  <div class="pb-fill" style="
                      background-color: #bdd2e6;
                      width: 0%;">
                    <style type="text/css" scoped="scoped">
                        @keyframes pb-fill-anim {{
                            0% {{ background-position: 0 0; }}
                            100% {{ background-position: 100px 0; }}
                        }}
                    </style>
                    &nbsp;
                  </div>
                </div>'''.format(uuid=self.uuid)

    def _js_update(self, progress):
        if progress is None:
            text = ''
        elif progress.finished:
            text = "{} finished in {}.".format(
                escape(progress.name_after),
                timestamp2timedelta(progress.elapsed_seconds()))
        elif progress.max_steps is None:
            text = (
                "{task}&hellip; duration: {duration}".format(
                    task=escape(progress.name_during),
                    duration=timestamp2timedelta(progress.elapsed_seconds())))
        else:
            text = (
                "{task}&hellip; {progress:.0f}%, ETA: {eta}".format(
                    task=escape(progress.name_during),
                    progress=100. * progress.progress,
                    eta=timestamp2timedelta(progress.eta())))

        if progress.max_steps is None:
            update = self._update_unknown_steps(progress)
        else:
            update = self._update_known_steps(progress)

        if progress.finished:
            finish = '''
                fill.style.animation = 'none';
                fill.style.backgroundImage = 'none';
            '''
        else:
            finish = ''

        return Javascript('''
              (function () {{
                  var root = document.getElementById('{uuid}');
                  var text = root.getElementsByClassName('pb-text')[0];
                  var fill = root.getElementsByClassName('pb-fill')[0];

                  text.innerHTML = '{text}';
                  {update}
                  {finish}
              }})();
        '''.format(uuid=self._uuid, text=text, update=update, finish=finish))

    def _update_known_steps(self, progress):
        return '''
            if ({progress} > 0.) {{
                fill.style.transition = 'width 0.1s linear';
            }} else {{
                fill.style.transition = 'none';
            }}

            fill.style.width = '{progress}%';
            fill.style.animation = 'none';
            fill.style.backgroundImage = 'none'
        '''.format(progress=100. * progress.progress)

    def _update_unknown_steps(self, progress):
        return '''
            fill.style.width = '100%';
            fill.style.animation = 'pb-fill-anim 2s linear infinite';
            fill.style.backgroundSize = '100px 100%';
            fill.style.backgroundImage = 'repeating-linear-gradient(' +
                '90deg, #bdd2e6, #edf2f8 40%, #bdd2e6 80%, #bdd2e6)';
        '''


class IPython5ProgressBar(ProgressBar):
    """ProgressBar for IPython>=5 environments.

    Provides a HTML representation, except for in a pure terminal IPython
    (i.e. not an IPython kernel that was connected to via ZMQ), where a
    ASCII progress bar will be used.

    Note that some Jupyter environments (like qtconsole) will try to use the
    HTML version, but do not support HTML and will show a warning instead of
    an actual progress bar.
    """
    supports_fast_ipynb_updates = True

    def __init__(self):
        super(IPython5ProgressBar, self).__init__()

        class Displayable(object):
            def __init__(self):
                self.display_requested = False

            def _ipython_display_(self):
                self.display_requested = True
        d = Displayable()
        display(d, exclude=['text/plain'])

        if d.display_requested:
            self._progress_bar = HtmlProgressBar()
        else:
            self._progress_bar = TerminalProgressBar()

    def update(self, progress):
        self._progress_bar.update(progress)


class WriteProgressToFile(ProgressBar):
    """Writes progress to a file.

    This is useful for remotely and intermittently monitoring progress.
    Note that this file will be overwritten on each update of the progress!

    Parameters
    ----------
    filename : str
        Path to the file to write the progress to.
    """

    def __init__(self, filename):
        self.filename = filename
        super(WriteProgressToFile, self).__init__()

    def update(self, progress):
        if progress.finished:
            text = "{} finished in {}.".format(
                self.progress.name_after,
                timestamp2timedelta(progress.elapsed_seconds()))
        else:
            text = "{progress:.0f}%, ETA: {eta}".format(
                progress=100 * progress.progress,
                eta=timestamp2timedelta(progress.eta()))

        with open(self.filename, 'w') as f:
            f.write(text + os.linesep)


class AutoProgressBar(ProgressBar):
    """Suppresses the progress bar unless the ETA exceeds a threshold.

    Parameters
    ----------
    delegate : :class:`ProgressBar`
        The actual progress bar to display, if ETA is high enough.
    min_eta : float, optional
        The minimum ETA threshold for displaying the progress bar.
    """

    def __init__(self, delegate, min_eta=1.):
        self.delegate = delegate

        super(AutoProgressBar, self).__init__()

        self.min_eta = min_eta
        self._visible = False

    def update(self, progress):
        min_delay = progress.time_start + 0.1
        long_eta = (progress.elapsed_seconds() + progress.eta() > self.min_eta
                    and min_delay < time.time())
        if self._visible:
            self.delegate.update(progress)
        elif long_eta or progress.finished:
            self._visible = True
            self.delegate.update(progress)

    @property
    def supports_fast_ipynb_updates(self):
        return self.delegate.supports_fast_ipynb_updates

    def close(self):
        self.delegate.close()


class ProgressTracker(object):
    """Tracks the progress of some process with a progress bar.

    Parameters
    ----------
    progress_bar : :class:`ProgressBar` or bool or None
        The progress bar to display the progress (or True to use the default
        progress bar, False/None to disable progress bar).
    total_progress : int
        Maximum number of steps of the process.
    update_interval : float, optional
        Time to wait (in seconds) between updates to progress bar display.
    """
    def __init__(self, progress_bar, total_progress, update_interval=0.1):
        self.progress_bar = to_progressbar(progress_bar)
        self.total_progress = total_progress
        self.update_interval = update_interval
        self.update_thread = threading.Thread(target=self.update_loop)
        self.update_thread.daemon = True
        self._closing = False
        self.sub_progress = None

    def next_stage(self, name_during='', name_after=None, max_steps=None):
        """Begin tracking progress of a new stage.

        Parameters
        ----------
        max_steps : int, optional
            The total number of calculation steps of the process.
        name_during : str, optional
            Short description of the task to be used while it is running.
        name_after : str, optional
            Short description of the task to be used after it has
            finished. Defaults to *name_during*.
        """
        if self.sub_progress is not None:
            self.total_progress.step()
        self.sub_progress = Progress(name_during, name_after, max_steps)
        return self.sub_progress

    def __enter__(self):
        self._closing = False
        self.total_progress.__enter__()
        if not isinstance(self.progress_bar, NoProgressBar):
            self.update_thread.start()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._closing = True
        self.total_progress.__exit__(exc_type, exc_value, traceback)
        if not isinstance(self.progress_bar, NoProgressBar):
            self.update_thread.join()
        self.progress_bar.update(self.total_progress)
        self.progress_bar.close()

    def update_loop(self):
        """Update the progress bar display (will run in a separate thread)."""

        while not self._closing:
            if (self.sub_progress is not None and
                    not self.sub_progress.finished):
                self.progress_bar.update(self.sub_progress)
            else:
                self.progress_bar.update(self.total_progress)
            time.sleep(self.update_interval)


def get_default_progressbar():
    """The default progress bar to use depending on the execution environment.

    Returns
    -------
    :class:`ProgressBar`
    """
    try:
        pbar = rc.getboolean('progress', 'progress_bar')
        if pbar:
            pbar = 'auto'
        else:
            pbar = 'none'
    except ValueError:
        pbar = rc.get('progress', 'progress_bar')

    if pbar.lower() == 'auto':
        if get_ipython() is not None and check_ipy_version((5, 0)):
            return AutoProgressBar(IPython5ProgressBar())
        else:
            return AutoProgressBar(TerminalProgressBar())
    if pbar.lower() == 'none':
        return NoProgressBar()

    try:
        return _load_class(pbar)()
    except Exception as e:
        warnings.warn(str(e))
        return NoProgressBar()


def to_progressbar(progress_bar):
    """Converts to a `.ProgressBar` instance.

    Parameters
    ----------
    progress_bar : None, bool, or `.ProgressBar`
        Object to be converted to a `.ProgressBar`.

    Returns
    -------
    ProgressBar
        Return *progress_bar* if it is already a progress bar, the default
        progress bar if *progress_bar* is *True*, and `NoProgressBar` if it is
        *None* or *False*.
    """
    if progress_bar is False or progress_bar is None:
        progress_bar = NoProgressBar()
    if progress_bar is True:
        progress_bar = get_default_progressbar()
    return progress_bar



import numpy as np
import pytest

import nengo


@pytest.mark.noassertions
@pytest.mark.parametrize('use_eventplot', [True, False])
def test_rasterplot(use_eventplot, Simulator, seed, plt):
    from nengo.utils.matplotlib import rasterplot

    with nengo.Network(seed=seed) as model:
        u = nengo.Node(output=lambda t: np.sin(6 * t))
        a = nengo.Ensemble(100, 1)
        nengo.Connection(u, a)
        ap = nengo.Probe(a.neurons)

    with Simulator(model) as sim:
        sim.run(1.0)

    rasterplot(sim.trange(), sim.data[ap], use_eventplot=use_eventplot)



import os

import numpy as np
from numpy.testing import assert_equal
import pytest

import nengo.utils.nco as nco
from nengo.utils.nco import Subfile


@pytest.fixture
def data():
    return '0123456789\n123456789'


@pytest.fixture
def testfile(data, tmpdir):
    f = tmpdir.join('file.txt')
    f.write(data)
    return f


class TestSubfile(object):
    def test_reads_only_from_start_to_end(self, data, testfile):
        with testfile.open() as f:
            assert Subfile(f, 2, 6).read() == data[2:6]
        with testfile.open() as f:
            assert Subfile(f, 2, 6).read(8) == data[2:6]
        with testfile.open() as f:
            assert Subfile(f, 2, 6).readline() == data[2:6]
        with testfile.open() as f:
            assert Subfile(f, 2, 6).readline(8) == data[2:6]

    def test_respects_read_size(self, data, testfile):
        with testfile.open() as f:
            assert Subfile(f, 2, 6).read(2) == data[2:4]
        with testfile.open() as f:
            assert Subfile(f, 2, 6).readline(2) == data[2:4]

    def test_readline(self, data, testfile):
        with testfile.open() as f:
            assert Subfile(f, 2, 14).readline() == data[2:11]
        with testfile.open() as f:
            assert Subfile(f, 2, 14).readline(15) == data[2:11]

    def test_seek(self, data, testfile):
        with testfile.open() as f:
            sf = Subfile(f, 2, 6)
            sf.seek(2)
            assert sf.read() == data[4:6]
        with testfile.open() as f:
            sf = Subfile(f, 2, 6)
            sf.read(1)
            sf.seek(2, os.SEEK_CUR)
            assert sf.read() == data[5:6]
        with testfile.open() as f:
            sf = Subfile(f, 2, 6)
            sf.seek(-2, os.SEEK_END)
            assert sf.read() == data[4:6]

    def test_seek_before_start(self, data, testfile):
        with testfile.open() as f:
            sf = Subfile(f, 2, 6)
            sf.seek(-2)
            assert sf.read() == data[2:6]
        with testfile.open() as f:
            sf = Subfile(f, 2, 6)
            sf.read(1)
            sf.seek(-4, os.SEEK_CUR)
            assert sf.read() == data[2:6]
        with testfile.open() as f:
            sf = Subfile(f, 2, 6)
            sf.seek(-8, os.SEEK_END)
            assert sf.read() == data[2:6]

    def test_seek_after_end(self, testfile):
        with testfile.open() as f:
            sf = Subfile(f, 2, 6)
            sf.seek(8)
            assert sf.read() == ''
        with testfile.open() as f:
            sf = Subfile(f, 2, 6)
            sf.read(1)
            sf.seek(8, os.SEEK_CUR)
            assert sf.read() == ''
        with testfile.open() as f:
            sf = Subfile(f, 2, 6)
            sf.seek(8, os.SEEK_END)
            assert sf.read() == ''


def test_nco_roundtrip(tmpdir):
    tmpfile = tmpdir.join('test.nco')

    pickle_data = {'0': 237, 'str': 'foobar'}
    array = np.array([[4, 3], [2, 1]])

    with tmpfile.open('wb') as f:
        nco.write(f, pickle_data, array)

    with tmpfile.open('rb') as f:
        pickle_data2, array2 = nco.read(f)

    assert pickle_data == pickle_data2
    assert_equal(array, array2)
